import time
import warnings
from functools import partialmethod
from typing import TYPE_CHECKING, AsyncGenerator, Dict, List, Optional, Type, Union

import grpc.aio

from jina.clients.base.grpc import GRPCBaseClient
from jina.clients.base.retry import wait_or_raise_err
from jina.excepts import InternalNetworkError
from jina.helper import deprecate_by, get_or_reuse_loop, run_async
from jina.importer import ImportExtensions

if TYPE_CHECKING:  # pragma: no cover
    from pydantic import BaseModel
    from jina.clients.base import CallbackFnType, InputType
    from jina.types.request.data import Response

from jina._docarray import Document, DocumentArray, docarray_v2


def _include_results_field_in_param(parameters: Optional['Dict']) -> 'Dict':
    key_result = '__results__'

    if parameters:

        if key_result in parameters:
            if not isinstance(parameters[key_result], dict):
                warnings.warn(
                    f'It looks like you passed a dictionary with the key `{key_result}` to `parameters`.'
                    'This key is reserved, so the associated value will be deleted.'
                )
                parameters.update({key_result: dict()})
    else:
        parameters = {key_result: dict()}

    return parameters


class MutateMixin:
    """The GraphQL Mutation Mixin for Client and Flow"""

    def mutate(
        self,
        mutation: str,
        variables: Optional[dict] = None,
        timeout: Optional[float] = None,
        headers: Optional[dict] = None,
    ):
        """Perform a GraphQL mutation

        :param mutation: the GraphQL mutation as a single string.
        :param variables: variables to be substituted in the mutation. Not needed if no variables are present in the mutation string.
        :param timeout: HTTP request timeout
        :param headers: HTTP headers
        :return: dict containing the optional keys ``data`` and ``errors``, for response data and errors.
        """
        with ImportExtensions(required=True):
            from sgqlc.endpoint.http import HTTPEndpoint as SgqlcHTTPEndpoint

            proto = 'https' if self.args.tls else 'http'
            graphql_url = f'{proto}://{self.args.host}:{self.args.port}/graphql'
            endpoint = SgqlcHTTPEndpoint(graphql_url)
            res = endpoint(
                mutation, variables=variables, timeout=timeout, extra_headers=headers
            )
            if 'errors' in res and res['errors']:
                msg = 'GraphQL mutation returned the following errors: '
                for err in res['errors']:
                    msg += err['message'] + '. '
                raise ConnectionError(msg)
            return res


class AsyncMutateMixin(MutateMixin):
    """The async GraphQL Mutation Mixin for Client and Flow"""

    async def mutate(
        self,
        mutation: str,
        variables: Optional[dict] = None,
        timeout: Optional[float] = None,
        headers: Optional[dict] = None,
    ):
        """Perform a GraphQL mutation, asynchronously

        :param mutation: the GraphQL mutation as a single string.
        :param variables: variables to be substituted in the mutation. Not needed if no variables are present in the mutation string.
        :param timeout: HTTP request timeout
        :param headers: HTTP headers
        :return: dict containing the optional keys ``data`` and ``errors``, for response data and errors.
        """
        return await get_or_reuse_loop().run_in_executor(
            None, super().mutate, mutation, variables, timeout, headers
        )


class HealthCheckMixin:
    """The Health check Mixin for Client and Flow to expose `dry_run` API"""

    def is_flow_ready(self, **kwargs) -> bool:
        """Check if the Flow is ready to receive requests

        :param kwargs: potential kwargs received passed from the public interface
        :return: boolean indicating the health/readiness of the Flow
        """
        return run_async(self.client._is_flow_ready, **kwargs)

    dry_run = deprecate_by(is_flow_ready)


class AsyncHealthCheckMixin:
    """The Health check Mixin for Client and Flow to expose `dry_run` API"""

    async def is_flow_ready(self, **kwargs) -> bool:
        """Check if the Flow is ready to receive requests

        :param kwargs: potential kwargs received passed from the public interface
        :return: boolean indicating the health/readiness of the Flow
        """
        return await self.client._is_flow_ready(**kwargs)

    dry_run = deprecate_by(is_flow_ready)


def _render_response_table(r, st, ed, show_table: bool = True):
    from rich import print

    elapsed = (ed - st) * 1000
    route = r.routes
    gateway_time = (
        route[0].end_time.ToMilliseconds() - route[0].start_time.ToMilliseconds()
    )
    exec_time = {}

    if len(route) > 1:
        for r in route[1:]:
            exec_time[r.executor] = (
                r.end_time.ToMilliseconds() - r.start_time.ToMilliseconds()
            )
    network_time = elapsed - gateway_time
    server_network = gateway_time - sum(exec_time.values())
    from rich.table import Table

    def make_table(_title, _time, _percent):
        table = Table(show_header=False, box=None)
        table.add_row(
            _title, f'[b]{_time:.0f}[/b]ms', f'[dim]{_percent * 100:.0f}%[/dim]'
        )
        return table

    from rich.tree import Tree

    t = Tree(make_table('Roundtrip', elapsed, 1))
    t.add(make_table('Client-server network', network_time, network_time / elapsed))
    t2 = t.add(make_table('Server', gateway_time, gateway_time / elapsed))
    t2.add(
        make_table(
            'Gateway-executors network', server_network, server_network / gateway_time
        )
    )
    for _name, _time in exec_time.items():
        t2.add(make_table(_name, _time, _time / gateway_time))

    if show_table:
        print(t)
    return {
        'Roundtrip': elapsed,
        'Client-server network': network_time,
        'Server': gateway_time,
        'Gateway-executors network': server_network,
        **exec_time,
    }


class ProfileMixin:
    """The Profile Mixin for Client and Flow to expose `profile` API"""

    def profiling(self, show_table: bool = True) -> Dict[str, float]:
        """Profiling a single query's roundtrip including network and computation latency. Results is summarized in a Dict.

        :param show_table: whether to show the table or not.
        :return: the latency report in a dict.
        """
        from docarray import Document

        st = time.perf_counter()
        r = self.client.post(on='/', inputs=Document(), return_responses=True)
        ed = time.perf_counter()
        return _render_response_table(r[0], st, ed, show_table=show_table)


class AsyncProfileMixin:
    """The Profile Mixin for Client and Flow to expose `profile` API"""

    async def profiling(self, show_table: bool = True) -> Dict[str, float]:
        """Profiling a single query's roundtrip including network and computation latency. Results is summarized in a Dict.

        :param show_table: whether to show the table or not.
        :return: the latency report in a dict.
        """
        from docarray import Document

        st = time.perf_counter()
        async for r in self.client.post(
            on='/', inputs=Document(), return_responses=True
        ):
            ed = time.perf_counter()
            return _render_response_table(r, st, ed, show_table=show_table)


class PostMixin:
    """The Post Mixin class for Client and Flow"""

    def _with_retry(
        self,
        func,
        inputs,
        on_done,
        on_error,
        on_always,
        exec_endpoint,
        target_executor,
        parameters,
        request_size,
        max_attempts,
        initial_backoff,
        max_backoff,
        backoff_multiplier,
        results_in_order,
        stream,
        prefetch,
        **kwargs,
    ):
        is_document_or_documentarray = isinstance(inputs, Document) or isinstance(
            inputs, DocumentArray
        )
        if (
            is_document_or_documentarray
            and isinstance(self.client, GRPCBaseClient)
            and max_attempts > 1
            and stream
        ):
            for attempt in range(1, max_attempts + 1):
                try:
                    return PostMixin._run_async(
                        backoff_multiplier,
                        exec_endpoint,
                        func,
                        initial_backoff,
                        inputs,
                        kwargs,
                        max_attempts,
                        max_backoff,
                        on_always,
                        on_done,
                        on_error,
                        parameters,
                        prefetch,
                        request_size,
                        results_in_order,
                        stream,
                        target_executor,
                    )
                except (
                    grpc.aio.AioRpcError,
                    InternalNetworkError,
                    ConnectionError,
                ) as err:
                    run_async(
                        wait_or_raise_err,
                        attempt=attempt,
                        err=err,
                        max_attempts=max_attempts,
                        backoff_multiplier=backoff_multiplier,
                        initial_backoff=initial_backoff,
                        max_backoff=max_backoff,
                    )
        else:
            return PostMixin._run_async(
                backoff_multiplier,
                exec_endpoint,
                func,
                initial_backoff,
                inputs,
                kwargs,
                max_attempts,
                max_backoff,
                on_always,
                on_done,
                on_error,
                parameters,
                prefetch,
                request_size,
                results_in_order,
                stream,
                target_executor,
            )

    @staticmethod
    def _run_async(
        backoff_multiplier,
        exec_endpoint,
        func,
        initial_backoff,
        inputs,
        kwargs,
        max_attempts,
        max_backoff,
        on_always,
        on_done,
        on_error,
        parameters,
        prefetch,
        request_size,
        results_in_order,
        stream,
        target_executor,
    ):
        return run_async(
            func,
            inputs=inputs,
            on_done=on_done,
            on_error=on_error,
            on_always=on_always,
            exec_endpoint=exec_endpoint,
            target_executor=target_executor,
            parameters=parameters,
            request_size=request_size,
            max_attempts=max_attempts,
            initial_backoff=initial_backoff,
            max_backoff=max_backoff,
            backoff_multiplier=backoff_multiplier,
            results_in_order=results_in_order,
            stream=stream,
            prefetch=prefetch,
            **kwargs,
        )

    def post(
        self,
        on: str,
        inputs: Optional['InputType'] = None,
        on_done: Optional['CallbackFnType'] = None,
        on_error: Optional['CallbackFnType'] = None,
        on_always: Optional['CallbackFnType'] = None,
        parameters: Union[Dict, 'BaseModel', None] = None,
        target_executor: Optional[str] = None,
        request_size: int = 100,
        show_progress: bool = False,
        continue_on_error: bool = False,
        return_responses: bool = False,
        max_attempts: int = 1,
        initial_backoff: float = 0.5,
        max_backoff: float = 2,
        backoff_multiplier: float = 1.5,
        results_in_order: bool = False,
        stream: bool = True,
        prefetch: Optional[int] = None,
        return_type: Type[DocumentArray] = DocumentArray,
        **kwargs,
    ) -> Optional[Union['DocumentArray', List['Response']]]:
        """Post a general data request to the Flow.

        :param inputs: input data which can be a DocList, a BaseDoc, an Iterable, a function which returns an Iterable.
        :param on: the endpoint which is invoked. All the functions in the executors decorated by `@requests(on=...)` with the same endpoint are invoked.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is either resolved or rejected.
        :param parameters: the parameters that will be sent to the executor, this can be a Dict or a Pydantic model
        :param target_executor: a regex string. Only matching Executors will process the request.
        :param request_size: the number of Documents per request. <=0 means all inputs in one request.
        :param show_progress: if set, client will show a progress bar on receiving every request.
        :param continue_on_error: if set, a Request that causes an error will be logged only without blocking the further requests.
        :param return_responses: if set to True, the result will come as Response and not as a `DocumentArray`
        :param max_attempts: Number of sending attempts, including the original request.
        :param initial_backoff: The first retry will happen with a delay of random(0, initial_backoff)
        :param max_backoff: The maximum accepted backoff after the exponential incremental delay
        :param backoff_multiplier: The n-th attempt will occur at random(0, min(initialBackoff*backoffMultiplier**(n-1), maxBackoff))
        :param results_in_order: return the results in the same order as the inputs
        :param stream: Applicable only to grpc client. If True, the requests are sent to the target using the gRPC streaming interface otherwise the gRPC unary interface will be used. The value is True by default.
        :param prefetch: How many Requests are processed from the Client at the same time. If not provided then Gateway prefetch value will be used.
        :param return_type: the DocList or BaseDoc type to be returned. By default, it is `DocumentArray`.
        :param kwargs: additional parameters
        :return: None or DocumentArray containing all response Documents

        .. warning::
            ``target_executor`` uses ``re.match`` for checking if the pattern is matched. ``target_executor=='foo'`` will match both deployments with the name ``foo`` and ``foo_what_ever_suffix``.
        """

        c = self.client
        c.show_progress = show_progress
        c.continue_on_error = continue_on_error

        parameters = _include_results_field_in_param(parameters)

        return_results = (on_always is None) and (on_done is None)

        async def _get_results(*args, **kwargs):
            is_singleton = False
            inferred_return_type = return_type
            if docarray_v2:
                from docarray import DocList
                if not issubclass(return_type, DocList):
                    is_singleton = True
                    inferred_return_type = DocList[return_type]
            result = [] if return_responses else inferred_return_type([])

            async for resp in c._get_results(*args, **kwargs):

                if return_results:
                    resp.document_array_cls = inferred_return_type
                    if return_responses:
                        result.append(resp)
                    else:
                        result.extend(resp.data.docs)
            if return_results:
                if not return_responses and is_singleton and len(result) == 1:
                    return result[0]
                else:
                    return result

        return self._with_retry(
            func=_get_results,
            inputs=inputs,
            on_done=on_done,
            on_error=on_error,
            on_always=on_always,
            exec_endpoint=on,
            target_executor=target_executor,
            parameters=parameters,
            request_size=request_size,
            max_attempts=max(max_attempts, 1),
            initial_backoff=initial_backoff,
            max_backoff=max_backoff,
            backoff_multiplier=backoff_multiplier,
            results_in_order=results_in_order,
            stream=stream,
            prefetch=prefetch,
            on=on,
            **kwargs,
        )

    # ONLY CRUD, for other request please use `.post`
    index = partialmethod(post, '/index')
    search = partialmethod(post, '/search')
    update = partialmethod(post, '/update')
    delete = partialmethod(post, '/delete')


class AsyncPostMixin:
    """The Async Post Mixin class for AsyncClient and AsyncFlow"""

    async def post(
        self,
        on: str,
        inputs: Optional['InputType'] = None,
        on_done: Optional['CallbackFnType'] = None,
        on_error: Optional['CallbackFnType'] = None,
        on_always: Optional['CallbackFnType'] = None,
        parameters: Union[Dict, 'BaseModel', None] = None,
        target_executor: Optional[str] = None,
        request_size: int = 100,
        show_progress: bool = False,
        continue_on_error: bool = False,
        return_responses: bool = False,
        max_attempts: int = 1,
        initial_backoff: float = 0.5,
        max_backoff: float = 2,
        backoff_multiplier: float = 1.5,
        results_in_order: bool = False,
        stream: bool = True,
        prefetch: Optional[int] = None,
        return_type: Type[DocumentArray] = DocumentArray,
        **kwargs,
    ) -> AsyncGenerator[None, Union['DocumentArray', 'Response']]:
        """Async Post a general data request to the Flow.

        :param inputs: input data which can be a DocList, a BaseDoc, an Iterable, a function which returns an Iterable.
        :param on: the endpoint which is invoked. All the functions in the executors decorated by `@requests(on=...)` with the same endpoint are invoked.
        :param on_done: the function to be called when the :class:`Request` object is resolved.
        :param on_error: the function to be called when the :class:`Request` object is rejected.
        :param on_always: the function to be called when the :class:`Request` object is either resolved or rejected.
        :param parameters: the parameters that will be sent to the executor, this can be a Dict or a Pydantic model
        :param target_executor: a regex string. Only matching Executors will process the request.
        :param request_size: the number of Documents per request. <=0 means all inputs in one request.
        :param show_progress: if set, client will show a progress bar on receiving every request.
        :param continue_on_error: if set, a Request that causes an error will be logged only without blocking the further requests.
        :param return_responses: if set to True, the result will come as Response and not as a `DocumentArray`
        :param max_attempts: Number of sending attempts, including the original request.
        :param initial_backoff: The first retry will happen with a delay of random(0, initial_backoff)
        :param max_backoff: The maximum accepted backoff after the exponential incremental delay
        :param backoff_multiplier: The n-th attempt will occur at random(0, min(initialBackoff*backoffMultiplier**(n-1), maxBackoff))
        :param results_in_order: return the results in the same order as the inputs
        :param stream: Applicable only to grpc client. If True, the requests are sent to the target using the gRPC streaming interface otherwise the gRPC unary interface will be used. The value is True by default.
        :param prefetch: How many Requests are processed from the Client at the same time. If not provided then Gateway prefetch value will be used.
        :param return_type: the DocList or BaseDoc type to be returned. By default, it is `DocumentArray`.
        :param kwargs: additional parameters, can be used to pass metadata or authentication information in the server call
        :yield: Response object

        .. warning::
            ``target_executor`` uses ``re.match`` for checking if the pattern is matched. ``target_executor=='foo'`` will match both deployments with the name ``foo`` and ``foo_what_ever_suffix``.
        """
        c = self.client
        c.show_progress = show_progress
        c.continue_on_error = continue_on_error

        parameters = _include_results_field_in_param(parameters)

        async for result in c._get_results(
            on=on,
            inputs=inputs,
            on_done=on_done,
            on_error=on_error,
            on_always=on_always,
            exec_endpoint=on,
            target_executor=target_executor,
            parameters=parameters,
            request_size=request_size,
            max_attempts=max(max_attempts, 1),
            initial_backoff=initial_backoff,
            max_backoff=max_backoff,
            backoff_multiplier=backoff_multiplier,
            results_in_order=results_in_order,
            stream=stream,
            prefetch=prefetch,
            return_type=return_type,
            **kwargs,
        ):
            is_singleton = False
            if docarray_v2:
                from docarray import DocList
                if issubclass(return_type, DocList):
                    result.document_array_cls = return_type
                else:
                    is_singleton = True
                    result.document_array_cls = DocList[return_type]
            if not return_responses:
                ret_docs = result.data.docs
                if is_singleton and len(ret_docs) == 1:
                    yield ret_docs[0]
                else:
                    yield ret_docs
            else:
                yield result

    async def stream_doc(
        self,
        on: str,
        inputs: Document,
        parameters: Optional[Dict] = None,
        return_type: Type[Document] = Document,
        timeout: Optional[int] = None,
        **kwargs,
    ) -> AsyncGenerator[None, 'Document']:
        """Send one document to a streaming endpoint and receive results asynchronisly, one Document at a time.

        :param inputs: input data which can be an Iterable, a function which returns an Iterable, or a single Document.
        :param on: the endpoint which is invoked. All the functions in the executors decorated by `@requests(on=...)` with the same endpoint are invoked.
        :param parameters: the kwargs that will be sent to the executor
        :param return_type: the DocumentArray type to be returned. By default, it is `DocumentArray`.
        :param timeout: Timeout for the client to remain connected to the server.
        :param kwargs: additional parameters, can be used to pass metadata or authentication information in the server call
        :yield: Document object
        """
        c = self.client
        parameters = _include_results_field_in_param(parameters)

        async for doc in c._get_streaming_results(
            on=on,
            inputs=inputs,
            exec_endpoint=on,
            parameters=parameters,
            return_type=return_type,
            timeout=timeout,
            **kwargs,
        ):
            yield doc

    # ONLY CRUD, for other request please use `.post`
    index = partialmethod(post, '/index')
    search = partialmethod(post, '/search')
    update = partialmethod(post, '/update')
    delete = partialmethod(post, '/delete')



import os
from typing import Dict, Optional

from hubble.executor.helper import is_valid_docker_uri, parse_hub_uri
from hubble.executor.hubio import HubIO

from jina.constants import (
    __default_composite_gateway__,
    __default_executor__,
    __default_grpc_gateway__,
    __default_http_gateway__,
    __default_websocket_gateway__,
    __dynamic_base_gateway_hubble__
)
from jina.enums import PodRoleType


def resolve_image_name(uses: Optional[str]):
    """Resolves the image name to be used instead of uses (resolving docker images)

    :param uses: image name

    :return: image name equivalent
    """
    if uses in [__default_http_gateway__,
                __default_websocket_gateway__,
                __default_grpc_gateway__,
                __default_composite_gateway__]:
        image_name = os.getenv(
            'JINA_GATEWAY_IMAGE', None
        )
        if image_name is None:
            image_name = get_image_name(__dynamic_base_gateway_hubble__)
    elif uses is not None and uses != __default_executor__:
        image_name = get_image_name(uses)
    else:
        image_name = os.getenv(
            'JINA_GATEWAY_IMAGE', None
        )
        if image_name is None:
            image_name = get_image_name(__dynamic_base_gateway_hubble__)

    return image_name


def get_image_name(uses: str) -> str:
    """The image can be provided in different formats by the user.
    This function converts it to an image name which can be understood by k8s.
    It uses the Hub api to get the image name and the latest tag on Docker Hub.

    If you don't want to rebuild image on Jina Hub,
    you can set `JINA_HUB_NO_IMAGE_REBUILD` environment variable.

    :param uses: image name

    :return: normalized image name
    """
    try:
        rebuild_image = 'JINA_HUB_NO_IMAGE_REBUILD' not in os.environ
        scheme, name, tag, secret = parse_hub_uri(uses)
        meta_data, _ = HubIO.fetch_meta(
            name, tag, secret=secret, rebuild_image=rebuild_image, force=True
        )
        image_name = meta_data.image_name
        return image_name
    except Exception:
        if uses.startswith('docker'):
            # docker:// is a valid requirement and user may want to put its own image
            return uses.replace('docker://', '')
        raise


def to_compatible_name(name: str) -> str:
    """Converts the deployment name to a valid name for K8s and docker compose.

    :param name: name of the deployment
    :return: compatible name
    """
    return name.replace('/', '-').replace('_', '-').lower()


def get_base_executor_version():
    """
    Get the version of jina to be used
    :return: the version tag
    """
    import requests

    try:
        from jina import __version__
        url = 'https://registry.hub.docker.com/v2/repositories/jinaai/jina/tags'
        result: Dict = requests.get(url, params={'name': __version__}).json()
        if result.get('count', 0) > 0:
            return __version__
        else:
            return 'master'
    except:
        return 'master'


def construct_runtime_container_args(cargs, uses_metas, uses_with, pod_type):
    """
    Construct a set of Namespace arguments into a list of arguments to pass to a container entrypoint
    :param cargs: The namespace arguments
    :param uses_metas: The uses_metas to override
    :param uses_with: The uses_with to override
    :param pod_type: The pod_type
    :return: Arguments to pass to container
    """
    import json

    from jina.helper import ArgNamespace
    from jina.parsers import set_pod_parser

    taboo = {
        'uses_with',
        'uses_metas',
        'volumes',
        'uses_before',
        'uses_after',
        'workspace_id',
        'noblock_on_start',
        'env',
        'env_from_secret',
        'image_pull_secrets',
    }

    if pod_type == PodRoleType.HEAD:
        taboo.add('uses')
        taboo.add('workspace')

    if pod_type in {PodRoleType.WORKER, PodRoleType.GATEWAY}:
        taboo.add('polling')

    non_defaults = ArgNamespace.get_non_defaults_args(
        cargs,
        set_pod_parser(),
        taboo=taboo,
    )
    _args = ArgNamespace.kwargs2list(non_defaults)
    container_args = ['executor'] + _args
    if uses_metas is not None:
        container_args.extend(['--uses-metas', json.dumps(uses_metas)])
    if uses_with is not None:
        container_args.extend(['--uses-with', json.dumps(uses_with)])
    container_args.append('--native')
    return container_args


def validate_uses(uses: str):
    """Validate uses argument

    :param uses: uses argument
    :return: boolean indicating whether is a valid uses to be used in K8s or docker compose
    """
    # Uses can be either None (not specified), default gateway class, default executor or docker image
    # None => deplyoment uses base container image and uses is determined inside container
    # default gateway class or default executor => deployment uses base container and sets uses in command
    # container images => deployment uses the specified container image and uses is defined by container
    if (
            uses is None
            or uses
            in [
        __default_http_gateway__,
        __default_websocket_gateway__,
        __default_grpc_gateway__,
        __default_composite_gateway__,
        __default_executor__,
    ]
            or uses.startswith('docker://')
    ):
        return True

    try:
        return is_valid_docker_uri(uses)
    except ValueError:
        return False



import asyncio
import copy
from typing import TYPE_CHECKING, AsyncGenerator, Callable, List, Optional, Tuple, Type

import grpc.aio

from jina._docarray import DocumentArray, docarray_v2
from jina.excepts import InternalNetworkError
from jina.helper import GATEWAY_NAME
from jina.logging.logger import JinaLogger
from jina.serve.networking import GrpcConnectionPool
from jina.serve.runtimes.gateway.graph.topology_graph import TopologyGraph
from jina.serve.runtimes.helper import _is_param_for_specific_executor
from jina.serve.runtimes.monitoring import MonitoringRequestMixin
from jina.serve.runtimes.worker.request_handling import WorkerRequestHandler

if TYPE_CHECKING:  # pragma: no cover
    from asyncio import Future

    from opentelemetry.metrics import Meter
    from prometheus_client import CollectorRegistry

    from jina.types.request import Request


class AsyncRequestResponseHandler(MonitoringRequestMixin):
    """
    Class that handles the requests arriving to the gateway and the result extracted from the requests future.

    :param metrics_registry: optional metrics registry for prometheus used if we need to expose metrics from the executor or from the data request handler
    :param runtime_name: optional runtime_name that will be registered during monitoring
    """

    def __init__(
        self,
        metrics_registry: Optional['CollectorRegistry'] = None,
        meter: Optional['Meter'] = None,
        runtime_name: Optional[str] = None,
        logger: Optional[JinaLogger] = None,
    ):
        super().__init__(metrics_registry, meter, runtime_name)
        self._endpoint_discovery_finished = False
        self._gathering_endpoints = False
        self.logger = logger or JinaLogger(self.__class__.__name__)

    def handle_request(
        self, graph: 'TopologyGraph', connection_pool: 'GrpcConnectionPool'
    ) -> Callable[['Request'], 'Tuple[Future, Optional[Future]]']:
        """
        Function that handles the requests arriving to the gateway. This will be passed to the streamer.

        :param graph: The TopologyGraph of the Flow.
        :param connection_pool: The connection pool to be used to send messages to specific nodes of the graph
        :return: Return a Function that given a Request will return a Future from where to extract the response
        """

        async def gather_endpoints(request_graph):
            if not self._endpoint_discovery_finished:
                self._gathering_endpoints = True
                try:
                    _ = await request_graph._get_all_endpoints(connection_pool)
                except InternalNetworkError as err:
                    err_code = err.code()
                    if err_code == grpc.StatusCode.UNAVAILABLE:
                        err._details = (
                            err.details()
                            + f' |Gateway: Communication error while gathering endpoints with deployment at address(es) {err.dest_addr}. Head or worker(s) may be down.'
                        )
                        raise err
                    else:
                        raise
                except Exception as exc:
                    self.logger.error(f' Error gathering endpoints: {exc}')
                    raise exc
                self._endpoint_discovery_finished = True

        def _handle_request(
            request: 'Request', return_type: Type[DocumentArray]
        ) -> 'Tuple[Future, Optional[Future]]':
            self._update_start_request_metrics(request)
            # important that the gateway needs to have an instance of the graph per request
            request_graph = copy.deepcopy(graph)
            responding_tasks = []
            floating_tasks = []
            endpoint = request.header.exec_endpoint
            r = request.routes.add()
            r.executor = 'gateway'
            r.start_time.GetCurrentTime()
            # If the request is targeting a specific deployment, we can send directly to the deployment instead of
            # querying the graph
            num_outgoing_nodes = len(request_graph.origin_nodes)
            has_specific_params = False
            request_input_parameters = request.parameters
            for key in request_input_parameters:
                if _is_param_for_specific_executor(key):
                    has_specific_params = True
                    break

            target_executor = request.header.target_executor
            # reset it in case we send to an external gateway
            request.header.target_executor = ''
            exec_endpoint = request.header.exec_endpoint
            gather_endpoints_task = None
            if not self._endpoint_discovery_finished and not self._gathering_endpoints:
                gather_endpoints_task = asyncio.create_task(
                    gather_endpoints(request_graph)
                )

            init_task = None
            request_doc_ids = []

            if graph.has_filter_conditions:
                if not docarray_v2:
                    request_doc_ids = request.data.docs[
                        :, 'id'
                    ]  # used to maintain order of docs that are filtered by executors
                else:
                    init_task = gather_endpoints_task
                    from docarray import DocList
                    from docarray.base_doc import AnyDoc

                    prev_doc_array_cls = request.data.document_array_cls
                    request.data.document_array_cls = DocList[AnyDoc]
                    request_doc_ids = request.data.docs.id
                    request.data._loaded_doc_array = None
                    request.data.document_array_cls = prev_doc_array_cls
            else:
                init_task = None

            for origin_node in request_graph.origin_nodes:
                leaf_tasks = origin_node.get_leaf_req_response_tasks(
                    connection_pool=connection_pool,
                    request_to_send=request,
                    previous_task=None,
                    endpoint=endpoint,
                    target_executor_pattern=target_executor or None,
                    request_input_parameters=request_input_parameters,
                    request_input_has_specific_params=has_specific_params,
                    copy_request_at_send=num_outgoing_nodes > 1 and has_specific_params,
                    init_task=init_task,
                    return_type=return_type,
                )
                # Every origin node returns a set of tasks that are the ones corresponding to the leafs of each of their
                # subtrees that unwrap all the previous tasks. It starts like a chain of waiting for tasks from previous
                # nodes
                responding_tasks.extend([task for ret, task in leaf_tasks if ret])
                floating_tasks.extend([task for ret, task in leaf_tasks if not ret])

            def _sort_response_docs(response):
                # sort response docs according to their order in the initial request
                def sort_by_request_order(doc):
                    if doc.id in request_doc_ids:
                        return request_doc_ids.index(doc.id)
                    else:
                        return len(request_doc_ids)  # put new/unknown docs at the end

                sorted_docs = sorted(response.data.docs, key=sort_by_request_order)
                response.data.docs = DocumentArray(sorted_docs)

            async def _process_results_at_end_gateway(
                tasks: List[asyncio.Task], request_graph: TopologyGraph
            ) -> asyncio.Future:
                try:
                    partial_responses = await asyncio.gather(*tasks)
                except Exception:
                    # update here failed request
                    self._update_end_failed_requests_metrics()
                    raise
                partial_responses, metadatas = zip(*partial_responses)
                filtered_partial_responses = list(
                    filter(lambda x: x is not None, partial_responses)
                )

                response = filtered_partial_responses[0]
                # JoanFM: to keep the docs_map feature, need to add the routes in the WorkerRuntime but clear it here
                # so that routes are properly done. not very clean but refactoring would be costly for such a small
                # thing, `docs_map` reuses routes potentially not in the best way but works for now
                for i in reversed(range(len(response.routes))):
                    if response.routes[i].executor != GATEWAY_NAME:
                        del response.routes[i]
                request_graph.add_routes(response)

                if graph.has_filter_conditions:
                    _sort_response_docs(response)

                collect_results = request_graph.collect_all_results()
                resp_params = response.parameters
                if len(collect_results) > 0:
                    resp_params[WorkerRequestHandler._KEY_RESULT] = collect_results
                    response.parameters = resp_params
                return response

            # In case of empty topologies
            if not responding_tasks:
                r.end_time.GetCurrentTime()
                future = asyncio.Future()
                future.set_result((request, {}))
                responding_tasks.append(future)

            return (
                asyncio.ensure_future(
                    _process_results_at_end_gateway(responding_tasks, request_graph)
                ),
                asyncio.ensure_future(asyncio.gather(*floating_tasks))
                if len(floating_tasks) > 0
                else None,
            )

        return _handle_request

    def handle_single_document_request(
        self, graph: 'TopologyGraph', connection_pool: 'GrpcConnectionPool'
    ) -> Callable[['Request', Type[DocumentArray]], 'AsyncGenerator']:
        """
        Function that handles the requests arriving to the gateway. This will be passed to the streamer.

        :param graph: The TopologyGraph of the Flow.
        :param connection_pool: The connection pool to be used to send messages to specific nodes of the graph
        :return: Return a Function that given a Request will return a Future from where to extract the response
        """

        async def _handle_request(
            request: 'Request', return_type: Type[DocumentArray] = DocumentArray
        ) -> 'Tuple[Future, Optional[Future]]':
            self._update_start_request_metrics(request)
            # important that the gateway needs to have an instance of the graph per request
            request_graph = copy.deepcopy(graph)
            r = request.routes.add()
            r.executor = 'gateway'
            r.start_time.GetCurrentTime()
            # If the request is targeting a specific deployment, we can send directly to the deployment instead of
            # querying the graph
            # reset it in case we send to an external gateway
            exec_endpoint = request.header.exec_endpoint

            node = request_graph.all_nodes[
                0
            ]  # this assumes there is only one Executor behind this Gateway
            async for resp in node.stream_single_doc(
                request=request,
                connection_pool=connection_pool,
                endpoint=exec_endpoint,
                return_type=return_type,
            ):
                yield resp

        return _handle_request

    def handle_result(self) -> Callable[['Request'], 'Request']:
        """
        Function that handles the result when extracted from the request future

        :return: Return a Function that returns a request to be returned to the client
        """

        def _handle_result(result: 'Request'):
            """
            Function that handles the result when extracted from the request future

            :param result: The result returned to the gateway. It extracts the request to be returned to the client
            :return: Returns a request to be returned to the client
            """
            for route in result.routes:
                if route.executor == GATEWAY_NAME:
                    route.end_time.GetCurrentTime()

            self._update_end_request_metrics(result)

            return result

        return _handle_result



from typing import TYPE_CHECKING

from jina.parsers.helper import _update_gateway_args

if TYPE_CHECKING:
    from argparse import Namespace


def deployment(args: 'Namespace'):
    """
    Start a Deployment

    :param args: arguments coming from the CLI.
    """
    from jina.orchestrate.deployments import Deployment

    if args.uses:
        dep = Deployment.load_config(args.uses)
        with dep:
            dep.block()
    else:
        raise ValueError('starting a Deployment from CLI requires a valid `--uses`')


def pod(args: 'Namespace'):
    """
    Start a Pod

    :param args: arguments coming from the CLI.
    """
    from jina.orchestrate.pods.factory import PodFactory

    try:
        with PodFactory.build_pod(args) as p:
            p.join()
    except KeyboardInterrupt:
        pass


def executor_native(args: 'Namespace'):
    """
    Starts an Executor in a WorkerRuntime

    :param args: arguments coming from the CLI.
    """
    from jina.serve.executors.run import run, run_stateful
    import multiprocessing
    from jina.jaml import JAML
    envs = {}
    envs.update(args.env or {})
    if not args.stateful:
        run(name=args.name,
            args=args,
            runtime_cls=args.runtime_cls,
            envs=envs,
            is_started=multiprocessing.Event(),
            is_signal_handlers_installed=multiprocessing.Event(),
            is_shutdown=multiprocessing.Event(),
            is_ready=multiprocessing.Event(),
            jaml_classes=JAML.registered_classes())
    else:
        run_stateful(name=args.name,
                     args=args,
                     runtime_cls=args.runtime_cls,
                     envs=envs)


def executor(args: 'Namespace'):
    """
    Starts an Executor in any Runtime

    :param args: arguments coming from the CLI.

    :returns: return the same as `pod` or `worker_runtime`
    """
    args.host = args.host[0]
    args.port_monitoring = args.port_monitoring[0]

    if args.native:
        return executor_native(args)
    else:
        return pod(args)


def gateway(args: 'Namespace'):
    """
    Start a Gateway Deployment

    :param args: arguments coming from the CLI.
    """
    from jina.serve.runtimes.gateway.request_handling import GatewayRequestHandler
    from jina.serve.runtimes.asyncio import AsyncNewLoopRuntime

    args.port_monitoring = args.port_monitoring[0]
    _update_gateway_args(args)

    with AsyncNewLoopRuntime(args, req_handler_cls=GatewayRequestHandler) as runtime:
        runtime.logger.info(f'Gateway started')
        runtime.run_forever()


def ping(args: 'Namespace'):
    """
    Check the connectivity of a Pod

    :param args: arguments coming from the CLI.
    """
    from jina.checker import NetworkChecker

    NetworkChecker(args)


def dryrun(args: 'Namespace'):
    """
    Check the health of a Flow

    :param args: arguments coming from the CLI.
    """
    from jina.checker import dry_run_checker

    dry_run_checker(args)


def client(args: 'Namespace'):
    """
    Start a client connects to the gateway

    :param args: arguments coming from the CLI.
    """
    from jina.clients import Client

    Client(args)


def export(args: 'Namespace'):
    """
    Export the API

    :param args: arguments coming from the CLI.
    """
    from jina import exporter

    getattr(exporter, f'export_{args.export.replace("-", "_")}')(args)


def flow(args: 'Namespace'):
    """
    Start a Flow from a YAML file or a docker image

    :param args: arguments coming from the CLI.
    """
    from jina import Flow

    if args.uses:
        f = Flow.load_config(args.uses)
        with f:
            f.block()
    else:
        raise ValueError('starting a Flow from CLI requires a valid `--uses`')


def hub(args: 'Namespace'):
    """
    Start a hub builder for push, pull
    :param args: arguments coming from the CLI.
    """
    from hubble.executor.hubio import HubIO

    getattr(HubIO(args), args.hub_cli)()


def new(args: 'Namespace'):
    """
    Create a new jina project
    :param args:  arguments coming from the CLI.
    """
    import os
    import shutil

    from jina.constants import __resources_path__

    if args.type == 'deployment':
        shutil.copytree(
            os.path.join(__resources_path__, 'project-template', 'deployment'), os.path.abspath(args.name)
        )
    else:
        shutil.copytree(
            os.path.join(__resources_path__, 'project-template', 'flow'), os.path.abspath(args.name)
        )


def help(args: 'Namespace'):
    """
    Lookup the usage of certain argument in Jina API.

    :param args: arguments coming from the CLI.
    """
    from jina_cli.lookup import lookup_and_print

    lookup_and_print(args.query.lower())


def auth(args: 'Namespace'):
    """
    Authenticate a user
    :param args: arguments coming from the CLI.
    """
    from hubble import api

    getattr(api, args.auth_cli.replace('-', '_'))(args)


def cloud(args: 'Namespace'):
    """
    Use jcloud (Jina Cloud) commands
    :param args: arguments coming from the CLI.
    """
    from jcloud import api

    getattr(api, args.jc_cli.replace('-', '_'))(args)



from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Union

from jina._docarray import docarray_v2
from jina.importer import ImportExtensions
from jina.types.request.data import DataRequest

if TYPE_CHECKING:
    from jina.logging.logger import JinaLogger

if docarray_v2:
    from docarray import BaseDoc, DocList


def get_fastapi_app(
    request_models_map: Dict,
    caller: Callable,
    logger: 'JinaLogger',
    cors: bool = False,
    **kwargs,
):
    """
    Get the app from FastAPI as the REST interface.

    :param request_models_map: Map describing the endpoints and its Pydantic models
    :param caller: Callable to be handled by the endpoints of the returned FastAPI app
    :param logger: Logger object
    :param cors: If set, a CORS middleware is added to FastAPI frontend to allow cross-origin access.
    :param kwargs: Extra kwargs to make it compatible with other methods
    :return: fastapi app
    """
    with ImportExtensions(required=True):
        import pydantic
        from fastapi import FastAPI, HTTPException, Request
        from fastapi.middleware.cors import CORSMiddleware
        from pydantic import BaseModel, Field
        from pydantic.config import BaseConfig, inherit_config

    import os

    from jina.proto import jina_pb2
    from jina.serve.runtimes.gateway.models import _to_camel_case

    if not docarray_v2:
        logger.warning('Only docarray v2 is supported with Sagemaker. ')
        return

    class Header(BaseModel):
        request_id: Optional[str] = Field(
            description='Request ID', example=os.urandom(16).hex()
        )

        class Config(BaseConfig):
            alias_generator = _to_camel_case
            allow_population_by_field_name = True

    class InnerConfig(BaseConfig):
        alias_generator = _to_camel_case
        allow_population_by_field_name = True

    app = FastAPI()

    if cors:
        app.add_middleware(
            CORSMiddleware,
            allow_origins=['*'],
            allow_credentials=True,
            allow_methods=['*'],
            allow_headers=['*'],
        )
        logger.warning('CORS is enabled. This service is accessible from any website!')

    def add_post_route(
        endpoint_path,
        input_model,
        output_model,
        input_doc_list_model=None,
        output_doc_list_model=None,
    ):
        from docarray.base_doc.docarray_response import DocArrayResponse

        app_kwargs = dict(
            path=f'/{endpoint_path.strip("/")}',
            methods=['POST'],
            summary=f'Endpoint {endpoint_path}',
            response_model=Union[output_model, List[output_model]],
            response_class=DocArrayResponse,
        )

        def is_valid_csv(content: str) -> bool:
            import csv
            from io import StringIO

            try:
                f = StringIO(content)
                reader = csv.DictReader(f)
                for _ in reader:
                    pass

                return True
            except Exception:
                return False

        async def process(body) -> output_model:
            req = DataRequest()
            if body.header is not None:
                req.header.request_id = body.header.request_id

            if body.parameters is not None:
                req.parameters = body.parameters
            req.header.exec_endpoint = endpoint_path
            req.document_array_cls = DocList[input_doc_model]

            data = body.data
            if isinstance(data, list):
                req.data.docs = DocList[input_doc_list_model](data)
            else:
                req.data.docs = DocList[input_doc_list_model]([data])
                if body.header is None:
                    req.header.request_id = req.docs[0].id

            resp = await caller(req)
            status = resp.header.status

            if status.code == jina_pb2.StatusProto.ERROR:
                raise HTTPException(status_code=499, detail=status.description)
            else:
                return output_model(data=resp.docs, parameters=resp.parameters)

        @app.api_route(**app_kwargs)
        async def post(request: Request):
            content_type = request.headers.get('content-type')
            if content_type == 'application/json':
                json_body = await request.json()
                return await process(input_model(**json_body))

            elif content_type in ('text/csv', 'application/csv'):
                import csv
                from io import StringIO

                bytes_body = await request.body()
                csv_body = bytes_body.decode('utf-8')
                if not is_valid_csv(csv_body):
                    raise HTTPException(
                        status_code=400,
                        detail='Invalid CSV input. Please check your input.',
                    )

                # NOTE: Sagemaker only supports csv files without header, so we enforce
                # the header by getting the field names from the input model.
                # This will also enforce the order of the fields in the csv file.
                # This also means, all fields in the input model must be present in the
                # csv file including the optional ones.
                # We also expect the csv file to have no quotes and use the escape char '\'
                field_names = [f for f in input_doc_list_model.__fields__]
                data = []
                for line in csv.reader(
                    StringIO(csv_body),
                    delimiter=',',
                    quoting=csv.QUOTE_NONE,
                    escapechar='\\',
                ):
                    if len(line) != len(field_names):
                        raise HTTPException(
                            status_code=400,
                            detail=f'Invalid CSV format. Line {line} doesn\'t match '
                            f'the expected field order {field_names}.',
                        )
                    data.append(input_doc_list_model(**dict(zip(field_names, line))))

                return await process(input_model(data=data))

            else:
                raise HTTPException(
                    status_code=400,
                    detail=f'Invalid content-type: {content_type}. '
                    f'Please use either application/json or text/csv.',
                )

    for endpoint, input_output_map in request_models_map.items():
        if endpoint != '_jina_dry_run_':
            input_doc_model = input_output_map['input']['model']
            output_doc_model = input_output_map['output']['model']
            parameters_model = input_output_map['parameters']['model'] or Optional[Dict]
            default_parameters = (
                ... if input_output_map['parameters']['model'] else None
            )

            _config = inherit_config(InnerConfig, BaseDoc.__config__)
            endpoint_input_model = pydantic.create_model(
                f'{endpoint.strip("/")}_input_model',
                data=(Union[List[input_doc_model], input_doc_model], ...),
                parameters=(parameters_model, default_parameters),
                header=(Optional[Header], None),
                __config__=_config,
            )

            endpoint_output_model = pydantic.create_model(
                f'{endpoint.strip("/")}_output_model',
                data=(Union[List[output_doc_model], output_doc_model], ...),
                parameters=(Optional[Dict], None),
                __config__=_config,
            )

            add_post_route(
                endpoint,
                input_model=endpoint_input_model,
                output_model=endpoint_output_model,
                input_doc_list_model=input_doc_model,
                output_doc_list_model=output_doc_model,
            )

    from jina.serve.runtimes.gateway.health_model import JinaHealthModel

    # `/ping` route is required by AWS Sagemaker
    @app.get(
        path='/ping',
        summary='Get the health of Jina Executor service',
        response_model=JinaHealthModel,
    )
    async def _executor_health():
        """
        Get the health of this Gateway service.
        .. # noqa: DAR201

        """
        return {}

    return app



import os
from typing import Optional

import grpc
from grpc import RpcError
from grpc_health.v1 import health, health_pb2, health_pb2_grpc
from grpc_reflection.v1alpha import reflection

from jina._docarray import docarray_v2
from jina.proto import jina_pb2, jina_pb2_grpc
from jina.serve.helper import get_server_side_grpc_options
from jina.serve.networking.utils import send_health_check_async, send_health_check_sync
from jina.serve.runtimes.servers import BaseServer


class GRPCServer(BaseServer):
    """GRPC Server implementation"""

    def __init__(
        self,
        grpc_server_options: Optional[dict] = None,
        ssl_keyfile: Optional[str] = None,
        ssl_certfile: Optional[str] = None,
        proxy: bool = False,
        **kwargs,
    ):
        """Initialize the gateway
        :param grpc_server_options: Dictionary of kwargs arguments that will be passed to the grpc server as options when starting the server, example : {'grpc.max_send_message_length': -1}
        :param ssl_keyfile: the path to the key file
        :param ssl_certfile: the path to the certificate file
        :param proxy: If set, respect the http_proxy and https_proxy environment variables, otherwise, it will unset these proxy variables before start. gRPC seems to prefer no proxy
        :param kwargs: keyword args
        """
        super().__init__(**kwargs)
        if not proxy and os.name != 'nt':
            os.unsetenv('http_proxy')
            os.unsetenv('https_proxy')

        self.grpc_server_options = grpc_server_options
        self.grpc_tracing_server_interceptors = self.aio_tracing_server_interceptors()
        self.ssl_keyfile = ssl_keyfile
        self.ssl_certfile = ssl_certfile
        self.health_servicer = health.aio.HealthServicer()

    async def setup_server(self):
        """
        setup GRPC server
        """
        self.logger.debug(f'Setting up GRPC server')
        if docarray_v2:
            from jina.serve.runtimes.gateway.request_handling import (
                GatewayRequestHandler,
            )

            if isinstance(self._request_handler, GatewayRequestHandler):
                await self._request_handler.streamer._get_endpoints_input_output_models(
                    is_cancel=self.is_cancel
                )
                self._request_handler.streamer._validate_flow_docarray_compatibility()

        self.server = grpc.aio.server(
            options=get_server_side_grpc_options(self.grpc_server_options),
            interceptors=self.grpc_tracing_server_interceptors,
        )

        jina_pb2_grpc.add_JinaRPCServicer_to_server(self._request_handler, self.server)

        jina_pb2_grpc.add_JinaSingleDataRequestRPCServicer_to_server(
            self._request_handler, self.server
        )

        if hasattr(self._request_handler, 'stream_doc'):
            jina_pb2_grpc.add_JinaSingleDocumentRequestRPCServicer_to_server(
                self._request_handler, self.server
            )
        if hasattr(self._request_handler, 'endpoint_discovery'):
            jina_pb2_grpc.add_JinaDiscoverEndpointsRPCServicer_to_server(
                self._request_handler, self.server
            )
        if hasattr(self._request_handler, 'process_data'):
            jina_pb2_grpc.add_JinaDataRequestRPCServicer_to_server(
                self._request_handler, self.server
            )
        if hasattr(self._request_handler, 'dry_run'):
            jina_pb2_grpc.add_JinaGatewayDryRunRPCServicer_to_server(
                self._request_handler, self.server
            )
        if hasattr(self._request_handler, 'snapshot'):
            jina_pb2_grpc.add_JinaExecutorSnapshotServicer_to_server(
                self._request_handler, self.server
            )
        if hasattr(self._request_handler, 'snapshot_status'):
            jina_pb2_grpc.add_JinaExecutorSnapshotProgressServicer_to_server(
                self._request_handler, self.server
            )
        if hasattr(self._request_handler, 'restore'):
            jina_pb2_grpc.add_JinaExecutorRestoreServicer_to_server(
                self._request_handler, self.server
            )
        if hasattr(self._request_handler, 'restore_status'):
            jina_pb2_grpc.add_JinaExecutorRestoreProgressServicer_to_server(
                self._request_handler, self.server
            )

        jina_pb2_grpc.add_JinaInfoRPCServicer_to_server(
            self._request_handler, self.server
        )

        service_names = (
            jina_pb2.DESCRIPTOR.services_by_name['JinaRPC'].full_name,
            jina_pb2.DESCRIPTOR.services_by_name['JinaSingleDataRequestRPC'].full_name,
            jina_pb2.DESCRIPTOR.services_by_name['JinaDataRequestRPC'].full_name,
            jina_pb2.DESCRIPTOR.services_by_name['JinaGatewayDryRunRPC'].full_name,
            jina_pb2.DESCRIPTOR.services_by_name['JinaSingleDocumentRequestRPC'].full_name,
            jina_pb2.DESCRIPTOR.services_by_name['JinaDiscoverEndpointsRPC'].full_name,
            jina_pb2.DESCRIPTOR.services_by_name['JinaInfoRPC'].full_name,
            reflection.SERVICE_NAME,
        )
        # Mark all services as healthy.
        health_pb2_grpc.add_HealthServicer_to_server(self.health_servicer, self.server)

        reflection.enable_server_reflection(service_names, self.server)

        bind_addr = f'{self.host}:{self.port}'

        if self.ssl_keyfile and self.ssl_certfile:
            with open(self.ssl_keyfile, 'rb') as f:
                private_key = f.read()
            with open(self.ssl_certfile, 'rb') as f:
                certificate_chain = f.read()

            server_credentials = grpc.ssl_server_credentials(
                (
                    (
                        private_key,
                        certificate_chain,
                    ),
                )
            )
            self.server.add_secure_port(bind_addr, server_credentials)
        elif (
            self.ssl_keyfile != self.ssl_certfile
        ):  # if we have only ssl_keyfile and not ssl_certfile or vice versa
            raise ValueError(
                f"you can't pass a ssl_keyfile without a ssl_certfile and vice versa"
            )
        else:
            self.server.add_insecure_port(bind_addr)
        self.logger.info(f'start server bound to {bind_addr}')
        await self.server.start()
        self.logger.debug(f'server bound to {bind_addr} started')
        for service in service_names:
            await self.health_servicer.set(
                service, health_pb2.HealthCheckResponse.SERVING
            )
        self.logger.debug(f'GRPC server setup successful')

    async def shutdown(self):
        """Free other resources allocated with the server, e.g, gateway object, ..."""
        self.logger.debug(f'Shutting down server')
        await super().shutdown()
        await self.health_servicer.enter_graceful_shutdown()
        await self._request_handler.close()  # allow pending requests to be processed
        await self.server.stop(grace=None)
        self.logger.debug(f'Server shutdown finished')

    async def run_server(self):
        """Run GRPC server forever"""
        await self.server.wait_for_termination()

    @staticmethod
    def is_ready(
        ctrl_address: str, timeout: float = 1.0, logger=None, **kwargs
    ) -> bool:
        """
        Check if status is ready.
        :param ctrl_address: the address where the control request needs to be sent
        :param timeout: timeout of the health check in seconds
        :param logger: JinaLogger to be used
        :param kwargs: extra keyword arguments
        :return: True if status is ready else False.
        """
        try:
            from grpc_health.v1 import health_pb2, health_pb2_grpc

            response = send_health_check_sync(ctrl_address, timeout=timeout)
            return (
                response.status == health_pb2.HealthCheckResponse.ServingStatus.SERVING
            )
        except RpcError as exc:
            if logger:
                logger.debug(f'Exception: {exc}')
            return False

    @staticmethod
    async def async_is_ready(
        ctrl_address: str, timeout: float = 1.0, logger=None, **kwargs
    ) -> bool:
        """
        Async Check if status is ready.
        :param ctrl_address: the address where the control request needs to be sent
        :param timeout: timeout of the health check in seconds
        :param logger: JinaLogger to be used
        :param kwargs: extra keyword arguments
        :return: True if status is ready else False.
        """
        try:
            from grpc_health.v1 import health_pb2, health_pb2_grpc

            response = await send_health_check_async(ctrl_address, timeout=timeout)
            return (
                response.status == health_pb2.HealthCheckResponse.ServingStatus.SERVING
            )
        except RpcError as exc:
            if logger:
                logger.debug(f'Exception: {exc}')
            return False



import asyncio
import copy
from typing import Any, List, TYPE_CHECKING

from jina.serve.runtimes.servers import BaseServer

if TYPE_CHECKING:
    from jina.logging.logger import JinaLogger


class CompositeBaseServer(BaseServer):
    """Composite Base Server implementation from which u can inherit a specific custom composite one"""
    servers: List['BaseServer']
    logger: 'JinaLogger'

    def __init__(
            self,
            **kwargs,
    ):
        """Initialize the gateway
        :param kwargs: keyword args
        """
        super().__init__(**kwargs)
        self._kwargs = kwargs

    @property
    def _server_kwargs(self):
        ret = []
        # ignore monitoring and tracing args since they are not copyable
        ignored_attrs = [
            'metrics_registry',
            'tracer_provider',
            'grpc_tracing_server_interceptors',
            'aio_tracing_client_interceptors',
            'tracing_client_interceptor',
        ]
        for port, protocol in zip(self.ports, self.protocols):
            # ignore monitoring and tracing args since they are not copyable
            runtime_args = self._deepcopy_with_ignore_attrs(
                self.runtime_args, ignored_attrs
            )
            runtime_args.port = port
            runtime_args.protocol = protocol
            server_kwargs = {k: v for k, v in self._kwargs.items() if k != 'runtime_args'}
            server_kwargs['runtime_args'] = dict(vars(runtime_args))
            server_kwargs['req_handler'] = self._request_handler
            ret.append(server_kwargs)

        return ret

    @staticmethod
    def _deepcopy_with_ignore_attrs(obj: Any, ignore_attrs: List[str]) -> Any:
        """Deep copy an object and ignore some attributes

        :param obj: the object to copy
        :param ignore_attrs: the attributes to ignore
        :return: the copied object
        """

        memo = {}
        for k in ignore_attrs:
            if hasattr(obj, k):
                memo[id(getattr(obj, k))] = None  # getattr(obj, k)

        return copy.deepcopy(obj, memo)

    async def setup_server(self):
        """
        setup servers inside CompositeServer
        """
        self.logger.debug(f'Setting up Composite server')
        tasks = []
        for server in self.servers:
            tasks.append(asyncio.create_task(server.setup_server()))

        await asyncio.gather(*tasks)
        self.logger.debug(f'Composite server setup successful')

    async def shutdown(self):
        """Free other resources allocated with the server, e.g, gateway object, ..."""
        self.logger.debug(f'Shutting down server')
        await super().shutdown()
        shutdown_tasks = []
        for server in self.servers:
            shutdown_tasks.append(asyncio.create_task(server.shutdown()))

        await asyncio.gather(*shutdown_tasks)
        self.logger.debug(f'Server shutdown finished')

    async def run_server(self):
        """Run servers inside CompositeServer forever"""
        run_server_tasks = []
        for server in self.servers:
            run_server_tasks.append(asyncio.create_task(server.run_server()))

        await asyncio.gather(*run_server_tasks)

    @property
    def _should_exit(self) -> bool:
        should_exit_values = [
            getattr(server, 'should_exit', True) for server in self.servers
        ]
        return all(should_exit_values)


class CompositeServer(CompositeBaseServer):
    """Composite Server implementation"""

    def __init__(
            self,
            **kwargs,
    ):
        """Initialize the gateway
        :param kwargs: keyword args
        """
        super().__init__(**kwargs)
        from jina.parsers.helper import _get_gateway_class

        self.servers: List[BaseServer] = []
        for server_kwargs in self._server_kwargs:
            server_cls = _get_gateway_class(server_kwargs['runtime_args']['protocol'],
                                            works_as_load_balancer=self.works_as_load_balancer)
            server = server_cls(**server_kwargs)
            self.servers.append(server)
        self.gateways = self.servers  # for backwards compatibility



import copy
from typing import Any, Dict, List, Optional, Tuple, Union

from jina._docarray import docarray_v2

_SPECIFIC_EXECUTOR_SEPARATOR = '__'


def _spit_key_and_executor_name(key_name: str) -> Tuple[str]:
    """Split a specific key into a key, name pair

    ex: 'key__my_executor' will be split into 'key', 'my_executor'

    :param key_name: key name of the param
    :return: return the split 'key', 'executor_name' for the key_name
    """
    key_split = key_name.split(_SPECIFIC_EXECUTOR_SEPARATOR)

    new_key_name = key_split.pop(-1)
    executor_name = ''.join(key_split)

    return new_key_name, executor_name


def _get_name_from_replicas_name(name: str) -> Tuple[str]:
    """return the original name without the replicas

    ex: 'exec1/rep-0' will be transform into 'exec1'

    :param name: name of the DataRequest
    :return: return the original name without the replicas
    """
    return name.split('/')[0]


def _is_param_for_specific_executor(key_name: str) -> bool:
    """Tell if a key is for a specific Executor

    ex: 'key' is for every Executor whereas 'my_executor__key' is only for 'my_executor'

    :param key_name: key name of the param
    :return: return True if key_name is for specific Executor, False otherwise
    """
    if _SPECIFIC_EXECUTOR_SEPARATOR in key_name:
        if key_name.startswith(_SPECIFIC_EXECUTOR_SEPARATOR) or key_name.endswith(
            _SPECIFIC_EXECUTOR_SEPARATOR
        ):
            return False
        return True
    else:
        return False


def _parse_specific_params(parameters: Dict, executor_name: str):
    """Parse the parameters dictionary to filter executor specific parameters

    :param parameters: dictionary container the parameters
    :param executor_name: name of the Executor
    :returns: the parsed parameters after applying filtering for the specific Executor
    """
    parsed_params = copy.deepcopy(parameters)

    for key in parameters:
        if _is_param_for_specific_executor(key):
            (
                key_name,
                key_executor_name,
            ) = _spit_key_and_executor_name(key)

            if key_executor_name == executor_name:
                parsed_params[key_name] = parameters[key]

            del parsed_params[key]

    specific_parameters = parameters.get(executor_name, None)
    if specific_parameters:
        parsed_params.update(**specific_parameters)

    return parsed_params


if docarray_v2:
    from docarray import BaseDoc, DocList
    from docarray.typing import AnyTensor
    from pydantic import create_model
    from pydantic.fields import FieldInfo

    from jina._docarray import docarray_v2

    RESERVED_KEYS = [
        'type',
        'anyOf',
        '$ref',
        'additionalProperties',
        'allOf',
        'items',
        'definitions',
        'properties',
        'default',
    ]

    def _create_aux_model_doc_list_to_list(model):
        fields: Dict[str, Any] = {}
        for field_name, field in model.__annotations__.items():
            if field_name not in model.__fields__:
                continue
            field_info = model.__fields__[field_name].field_info
            try:
                if issubclass(field, DocList):
                    t: Any = field.doc_type
                    t_aux = _create_aux_model_doc_list_to_list(t)
                    fields[field_name] = (List[t_aux], field_info)
                else:
                    fields[field_name] = (field, field_info)
            except TypeError:
                fields[field_name] = (field, field_info)
        return create_model(
            model.__name__,
            __base__=model,
            __validators__=model.__validators__,
            **fields,
        )

    def _get_field_from_type(
        field_schema,
        field_name,
        root_schema,
        cached_models,
        is_tensor=False,
        num_recursions=0,
        base_class=BaseDoc,
        definitions: Optional[Dict] = None,
    ):
        if not definitions:
            definitions = {}
        field_type = field_schema.get('type', None)
        tensor_shape = field_schema.get('tensor/array shape', None)
        if 'anyOf' in field_schema:
            any_of_types = []
            for any_of_schema in field_schema['anyOf']:
                if '$ref' in any_of_schema:
                    obj_ref = any_of_schema.get('$ref')
                    ref_name = obj_ref.split('/')[-1]
                    any_of_types.append(
                        _create_pydantic_model_from_schema(
                            definitions[ref_name],
                            ref_name,
                            cached_models=cached_models,
                            base_class=base_class,
                            definitions=definitions,
                        )
                    )
                else:
                    any_of_types.append(
                        _get_field_from_type(
                            any_of_schema,
                            field_name,
                            root_schema=root_schema,
                            cached_models=cached_models,
                            is_tensor=tensor_shape is not None,
                            num_recursions=0,
                            base_class=base_class,
                            definitions=definitions,
                        )
                    )  # No Union of Lists
            ret = Union[tuple(any_of_types)]
            for rec in range(num_recursions):
                ret = List[ret]
        elif field_type == 'string':
            ret = str
            for rec in range(num_recursions):
                ret = List[ret]
        elif field_type == 'integer':
            ret = int
            for rec in range(num_recursions):
                ret = List[ret]
        elif field_type == 'number':
            if num_recursions <= 1:
                # This is a hack because AnyTensor is more generic than a simple List and it comes as simple List
                if is_tensor:
                    ret = AnyTensor
                else:
                    ret = List[float]
            else:
                ret = float
                for rec in range(num_recursions):
                    ret = List[ret]
        elif field_type == 'boolean':
            ret = bool
            for rec in range(num_recursions):
                ret = List[ret]
        elif field_type == 'object' or field_type is None:
            if 'additionalProperties' in field_schema:  # handle Dictionaries
                additional_props = field_schema['additionalProperties']
                if additional_props.get('type') == 'object':
                    ret = Dict[
                        str,
                        _create_pydantic_model_from_schema(
                            additional_props,
                            field_name,
                            cached_models=cached_models,
                            base_class=base_class,
                        ),
                    ]
                else:
                    ret = Dict[str, Any]
            else:
                obj_ref = field_schema.get('$ref') or field_schema.get('allOf', [{}])[
                    0
                ].get('$ref', None)
                if num_recursions == 0:  # single object reference
                    if obj_ref:
                        ref_name = obj_ref.split('/')[-1]
                        ret = _create_pydantic_model_from_schema(
                            definitions[ref_name],
                            ref_name,
                            cached_models=cached_models,
                            base_class=base_class,
                            definitions=definitions,
                        )
                    else:
                        ret = Any
                else:  # object reference in definitions
                    if obj_ref:
                        ref_name = obj_ref.split('/')[-1]
                        ret = DocList[
                            _create_pydantic_model_from_schema(
                                definitions[ref_name],
                                ref_name,
                                cached_models=cached_models,
                                base_class=base_class,
                                definitions=definitions,
                            )
                        ]
                    else:
                        ret = DocList[
                            _create_pydantic_model_from_schema(
                                field_schema,
                                field_name,
                                cached_models=cached_models,
                                base_class=base_class,
                                definitions=definitions,
                            )
                        ]
        elif field_type == 'array':
            ret = _get_field_from_type(
                field_schema=field_schema.get('items', {}),
                field_name=field_name,
                root_schema=root_schema,
                cached_models=cached_models,
                is_tensor=tensor_shape is not None,
                num_recursions=num_recursions + 1,
                base_class=base_class,
                definitions=definitions,
            )
        else:
            if num_recursions > 0:
                raise ValueError(
                    f"Unknown array item type: {field_type} for field_name {field_name}"
                )
            else:
                raise ValueError(
                    f"Unknown field type: {field_type} for field_name {field_name}"
                )
        return ret

    def _create_pydantic_model_from_schema(
        schema: Dict[str, any],
        model_name: str,
        cached_models: Dict,
        base_class=BaseDoc,
        definitions: Optional[Dict] = None,
    ) -> type:
        if not definitions:
            definitions = schema.get('definitions', {})
        cached_models = cached_models if cached_models is not None else {}
        fields: Dict[str, Any] = {}
        if model_name in cached_models:
            return cached_models[model_name]
        for field_name, field_schema in schema.get('properties', {}).items():
            field_type = _get_field_from_type(
                field_schema=field_schema,
                field_name=field_name,
                root_schema=schema,
                cached_models=cached_models,
                is_tensor=False,
                num_recursions=0,
                base_class=base_class,
                definitions=definitions,
            )
            fields[field_name] = (
                field_type,
                FieldInfo(default=field_schema.pop('default', None), **field_schema),
            )

        model = create_model(model_name, __base__=base_class, **fields)
        model.__config__.title = schema.get('title', model.__config__.title)

        for k in RESERVED_KEYS:
            if k in schema:
                schema.pop(k)
        model.__config__.schema_extra = schema
        cached_models[model_name] = model
        return model



import asyncio
import json
import threading
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type

import grpc
from grpc import RpcError

from jina._docarray import Document, docarray_v2
from jina.clients.base import BaseClient
from jina.clients.base.stream_rpc import StreamRpc
from jina.clients.base.unary_rpc import UnaryRpc
from jina.clients.helper import callback_exec
from jina.excepts import BadClientInput, BadServerFlow, InternalNetworkError
from jina.logging.profile import ProgressBar
from jina.proto import jina_pb2, jina_pb2_grpc
from jina.serve.helper import extract_trailing_metadata, get_default_grpc_options
from jina.serve.networking.utils import get_grpc_channel
from jina.types.request.data import SingleDocumentRequest

if TYPE_CHECKING:  # pragma: no cover
    from jina.clients.base import CallbackFnType, InputType


class GRPCBaseClient(BaseClient):
    """A simple Python client for connecting to the gRPC gateway.

    It manages the asyncio event loop internally, so all interfaces are synchronous from the outside.
    """

    _lock = threading.RLock()

    async def _is_flow_ready(self, **kwargs) -> bool:
        """Sends a dry run to the Flow to validate if the Flow is ready to receive requests

        :param kwargs: potential kwargs received passed from the public interface
        :return: boolean indicating the health/readiness of the Flow
        """
        try:
            async with get_grpc_channel(
                f'{self.args.host}:{self.args.port}',
                asyncio=True,
                tls=self.args.tls,
            ) as channel:
                stub = jina_pb2_grpc.JinaGatewayDryRunRPCStub(channel)
                self.logger.debug(f'connected to {self.args.host}:{self.args.port}')
                call_result = stub.dry_run(
                    jina_pb2.google_dot_protobuf_dot_empty__pb2.Empty(),
                    metadata=kwargs.get('metadata', None),
                    credentials=kwargs.get('credentials', None),
                    timeout=kwargs.get('timeout', None),
                )
                metadata, response = (
                    await call_result.trailing_metadata(),
                    await call_result,
                )
                if response.code == jina_pb2.StatusProto.SUCCESS:
                    return True
                else:
                    self.logger.error(
                        f'Returned code is not expected! Exception: {response.exception}'
                    )
        except RpcError as e:
            self.logger.error(f'RpcError: {e.details()}')
        except Exception as e:
            self.logger.error(f'Error while getting response from grpc server {e!r}')

        return False

    async def _get_results(
        self,
        inputs: 'InputType',
        on_done: 'CallbackFnType',
        on_error: Optional['CallbackFnType'] = None,
        on_always: Optional['CallbackFnType'] = None,
        compression: Optional[str] = None,
        max_attempts: int = 1,
        initial_backoff: float = 0.5,
        max_backoff: float = 2,
        backoff_multiplier: float = 1.5,
        results_in_order: bool = False,
        stream: bool = True,
        prefetch: Optional[int] = None,
        **kwargs,
    ):
        try:
            self.compression = (
                getattr(grpc.Compression, compression)
                if compression
                else grpc.Compression.NoCompression
            )

            self.inputs = inputs
            req_iter = self._get_requests(**kwargs)
            continue_on_error = self.continue_on_error
            # while loop with retries, check in which state the `iterator` remains after failure
            options = client_grpc_options(
                backoff_multiplier,
                initial_backoff,
                max_attempts,
                max_backoff,
                self.args.grpc_channel_options,
            )

            metadata = kwargs.pop('metadata', ())
            if results_in_order:
                metadata = metadata + (('__results_in_order__', 'true'),)

            if prefetch:
                metadata = metadata + (('__prefetch__', str(prefetch)),)

            with self._lock:
                async with get_grpc_channel(
                    f'{self.args.host}:{self.args.port}',
                    options=options,
                    asyncio=True,
                    tls=self.args.tls,
                    aio_tracing_client_interceptors=self.aio_tracing_client_interceptors(),
                ) as channel:
                    self.logger.debug(f'connected to {self.args.host}:{self.args.port}')

                    with ProgressBar(
                        total_length=self._inputs_length, disable=not self.show_progress
                    ) as p_bar:
                        try:
                            if stream:
                                stream_rpc = StreamRpc(
                                    channel=channel,
                                    continue_on_error=continue_on_error,
                                    metadata=metadata,
                                    on_always=on_always,
                                    on_done=on_done,
                                    on_error=on_error,
                                    p_bar=p_bar,
                                    req_iter=req_iter,
                                    max_attempts=max_attempts,
                                    backoff_multiplier=backoff_multiplier,
                                    initial_backoff=initial_backoff,
                                    max_backoff=max_backoff,
                                    logger=self.logger,
                                    show_progress=self.show_progress,
                                    compression=self.compression,
                                    **kwargs,
                                )
                                async for response in stream_rpc.stream_rpc_with_retry():
                                    yield response
                            else:
                                unary_rpc = UnaryRpc(
                                    channel=channel,
                                    continue_on_error=continue_on_error,
                                    metadata=metadata,
                                    on_always=on_always,
                                    on_done=on_done,
                                    on_error=on_error,
                                    p_bar=p_bar,
                                    req_iter=req_iter,
                                    max_attempts=max_attempts,
                                    backoff_multiplier=backoff_multiplier,
                                    initial_backoff=initial_backoff,
                                    max_backoff=max_backoff,
                                    logger=self.logger,
                                    show_progress=self.show_progress,
                                    compression=self.compression,
                                    client_args=self.args,
                                    prefetch=prefetch,
                                    results_in_order=results_in_order,
                                    **kwargs,
                                )
                                async for response in unary_rpc.unary_rpc_with_retry():
                                    yield response
                        except (grpc.aio.AioRpcError, InternalNetworkError) as err:
                            await self._handle_error_and_metadata(err)
        except KeyboardInterrupt:
            self.logger.warning('user cancel the process')
        except asyncio.CancelledError as ex:
            self.logger.warning(f'process error: {ex!r}')
            raise
        except:
            # Not sure why, adding this line helps in fixing a hanging test
            raise

    async def _handle_error_and_metadata(self, err):
        my_code = err.code()
        my_details = err.details()
        trailing_metadata = extract_trailing_metadata(err)
        msg = f'gRPC error: {my_code} {my_details}\n{trailing_metadata}'
        if my_code == grpc.StatusCode.UNAVAILABLE:
            self.logger.error(
                f'{msg}\nThe ongoing request is terminated as the server is not available or closed already.'
            )
            raise ConnectionError(msg)
        if my_code == grpc.StatusCode.NOT_FOUND:
            self.logger.error(
                f'{msg}\nThe ongoing request is terminated as a resource cannot be found.'
            )
            raise ConnectionError(msg)
        elif my_code == grpc.StatusCode.DEADLINE_EXCEEDED:
            self.logger.error(
                f'{msg}\nThe ongoing request is terminated due to a server-side timeout.'
            )
            raise ConnectionError(msg)
        elif my_code == grpc.StatusCode.INTERNAL:
            self.logger.error(f'{msg}\ninternal error on the server side')
            raise err
        elif (
            my_code == grpc.StatusCode.UNKNOWN
            and 'asyncio.exceptions.TimeoutError' in my_details
        ):
            raise BadClientInput(
                f'{msg}\n'
                'often the case is that you define/send a bad input iterator to jina, '
                'please double check your input iterator'
            ) from err
        else:
            raise BadServerFlow(msg) from err

    async def stream_doc_endpoint(
        self,
        request: SingleDocumentRequest,
        timeout: Optional[float] = None,
    ):
        """
        Use the stream_doc stub to send one document and stream documents back from the Executor

        :param request: The request to be sent
        :param timeout: defines timeout for sending request

        :yields: response document
        """
        async with get_grpc_channel(
            f'{self.args.host}:{self.args.port}',
            asyncio=True,
            tls=self.args.tls,
            aio_tracing_client_interceptors=self.aio_tracing_client_interceptors(),
        ) as channel:
            stub = jina_pb2_grpc.JinaSingleDocumentRequestRPCStub(channel)
            try:
                async for response in stub.stream_doc(request, timeout=timeout):
                    callback_exec(
                        response=response,
                        logger=self.logger,
                    )
                    yield response
            except (grpc.aio.AioRpcError, InternalNetworkError) as err:
                await self._handle_error_and_metadata(err)

    async def _get_streaming_results(
        self,
        on: str,
        inputs: 'Document',
        parameters: Optional[Dict] = None,
        return_type: Type[Document] = Document,
        timeout: Optional[int] = None,
        **kwargs,
    ):
        req = SingleDocumentRequest()
        req.header.exec_endpoint = on
        req.document_cls = inputs.__class__
        req.data.doc = inputs
        async for response in self.stream_doc_endpoint(
            request=req, timeout=timeout
        ):
            yield return_type.from_protobuf(response.document)


def client_grpc_options(
    backoff_multiplier: float,
    initial_backoff: float,
    max_attempts: int,
    max_backoff: float,
    args_channel_options: Optional[Dict[str, Any]],
) -> List[Tuple[str, Any]]:
    """
    Builds grpc options for the client by taking into account the retry parameters, default gRPC options and additional gRPC options.
    :param max_attempts: Maximum number of attempts that are allowed.
    :param backoff_multiplier: Factor that will be raised to the exponent of (attempt - 1) for calculating the backoff wait time.
    :param initial_backoff: The backoff time on the first error. This will be multiplied by the backoff_multiplier exponent for subsequent wait time calculations.
    :param max_backoff: The maximum backoff wait time.
    :param args_channel_options: additional gRPC options that must be merged with the default options
    :return: List of gRPC options
    """
    grpc_options = get_default_grpc_options()

    if args_channel_options:
        grpc_options.extend(list(args_channel_options))

    if max_attempts > 1:
        options_with_retry_config = remove_grpc_service_config(grpc_options)
        service_config_json = json.dumps(
            {
                "methodConfig": [
                    {
                        # To apply retry to all methods, put [{}] in the "name" field
                        "name": [{}],
                        "retryPolicy": {
                            "maxAttempts": max_attempts,
                            "initialBackoff": f"{initial_backoff}s",
                            "maxBackoff": f"{max_backoff}s",
                            "backoffMultiplier": backoff_multiplier,
                            "retryableStatusCodes": [
                                "UNAVAILABLE",
                                "DEADLINE_EXCEEDED",
                                "INTERNAL",
                            ],
                        },
                    }
                ]
            }
        )
        # NOTE: the retry feature will be enabled by default >=v1.40.0
        options_with_retry_config.append(("grpc.enable_retries", 1))
        options_with_retry_config.append(("grpc.service_config", service_config_json))
        return options_with_retry_config

    return grpc_options


def remove_grpc_service_config(options: List[Tuple[str, Any]]) -> List[Tuple[str, Any]]:
    """
    Removes the 'grpc.service_config' configuration from the list of options.
    :param options: List of gRPC options
    :return: List of gRPC options
    """
    return [tup for tup in options if tup[0] != 'grpc.service_config']



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: add_voter.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x61\x64\x64_voter.proto\"-\n\rAwaitResponse\x12\r\n\x05\x65rror\x18\x01 \x01(\t\x12\r\n\x05index\x18\x02 \x01(\x04\"\x10\n\x0e\x46orgetResponse\"!\n\x06\x46uture\x12\x17\n\x0foperation_token\x18\x01 \x01(\t\"F\n\x0f\x41\x64\x64VoterRequest\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12\x16\n\x0eprevious_index\x18\x03 \x01(\x04\x32~\n\tRaftAdmin\x12\'\n\x08\x41\x64\x64Voter\x12\x10.AddVoterRequest\x1a\x07.Future\"\x00\x12\"\n\x05\x41wait\x12\x07.Future\x1a\x0e.AwaitResponse\"\x00\x12$\n\x06\x46orget\x12\x07.Future\x1a\x0f.ForgetResponse\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'add_voter_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _AWAITRESPONSE._serialized_start=19
  _AWAITRESPONSE._serialized_end=64
  _FORGETRESPONSE._serialized_start=66
  _FORGETRESPONSE._serialized_end=82
  _FUTURE._serialized_start=84
  _FUTURE._serialized_end=117
  _ADDVOTERREQUEST._serialized_start=119
  _ADDVOTERREQUEST._serialized_end=189
  _RAFTADMIN._serialized_start=191
  _RAFTADMIN._serialized_end=317
# @@protoc_insertion_point(module_scope)



"""Extensive argparser module for logging"""


def mixin_suppress_root_logging_parser(parser):
    """Mixing in arguments required by every module into the given parser.
    This parser extends logging-related arguments.
    :param parser: the parser instance to which we add arguments
    """
    parser.add_argument(
        '--suppress-root-logging',
        action='store_true',
        default=False,
        help='If set, then no root handlers will be suppressed from logging.',
    )



from argparse import Namespace
from copy import deepcopy
from typing import TYPE_CHECKING, Type

from hubble.executor.helper import is_valid_huburi
from hubble.executor.hubio import HubIO

from jina.enums import PodRoleType
from jina.orchestrate.pods import Pod
from jina.orchestrate.pods.container import ContainerPod

if TYPE_CHECKING:  # pragma: no cover
    from jina.orchestrate.pods import BasePod


class PodFactory:
    """
    A PodFactory is a factory class, abstracting the Pod creation
    """

    @staticmethod
    def build_pod(args: 'Namespace', gateway_load_balancer: bool = False) -> Type['BasePod']:
        """Build an implementation of a `BasePod` interface

        :param args: deployment arguments parsed from the CLI.
        :param gateway_load_balancer: flag indicating if this Pod is supposed to be a Gateway Load Balancer

        :return: the created Deployment
        """
        # copy to update but forward original

        cargs = deepcopy(args)
        cargs.gateway_load_balancer = gateway_load_balancer

        if is_valid_huburi(cargs.uses):
            _hub_args = deepcopy(args)
            _hub_args.uri = args.uses
            _hub_args.no_usage = True
            cargs.uses = HubIO(_hub_args).pull()

        if (
            cargs.pod_role != PodRoleType.HEAD
            and cargs.uses
            and cargs.uses.startswith('docker://')
        ):
            return ContainerPod(cargs)
        else:
            return Pod(cargs)



"""Module for helper functions in the parser"""

import argparse
import os
from typing import Tuple

from jina.enums import ProtocolType
from jina.logging.predefined import default_logger

_SHOW_ALL_ARGS = 'JINA_FULL_CLI' in os.environ


def add_arg_group(parser, title):
    """Add the arguments for a specific group to the parser

    :param parser: the parser configure
    :param title: the group name
    :return: the new parser
    """
    return parser.add_argument_group(f'{title} arguments')


class KVAppendAction(argparse.Action):
    """argparse action to split an argument into KEY=VALUE form
    on the first = and append to a dictionary.
    This is used for setting up --env
    """

    def __call__(self, parser, args, values, option_string=None):
        """
        call the KVAppendAction


        .. # noqa: DAR401
        :param parser: the parser
        :param args: args to initialize the values
        :param values: the values to add to the parser
        :param option_string: inherited, not used
        """
        import json
        import re

        from jina.helper import parse_arg

        d = getattr(args, self.dest) or {}
        for value in values:
            try:
                d.update(json.loads(value))
            except json.JSONDecodeError:
                try:
                    k, v = re.split(r'[:=]\s*', value, maxsplit=1)
                except ValueError:
                    raise argparse.ArgumentTypeError(
                        f'could not parse argument \"{values[0]}\" as k=v format'
                    )
                d[k] = parse_arg(v)
        setattr(args, self.dest, d)


class _ColoredHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    class _Section(object):
        def __init__(self, formatter, parent, heading=None):
            self.formatter = formatter
            self.parent = parent
            self.heading = heading
            self.items = []

        def format_help(self):
            # format the indented section
            if self.parent is not None:
                self.formatter._indent()
            join = self.formatter._join_parts
            item_help = join([func(*args) for func, args in self.items])
            if self.parent is not None:
                self.formatter._dedent()

            # return nothing if the section was empty
            if not item_help.strip():
                return ''

            # add the heading if the section was non-empty
            if self.heading is not argparse.SUPPRESS and self.heading is not None:
                from jina.helper import colored

                current_indent = self.formatter._current_indent
                captial_heading = ' '.join(
                    v[0].upper() + v[1:] for v in self.heading.split(' ')
                )
                heading = '%*s%s\n' % (
                    current_indent,
                    '',
                    colored(f'▮ {captial_heading}', 'cyan', attrs=['bold']),
                )
            else:
                heading = ''

            # join the section-initial newline, the heading and the help
            return join(['\n', heading, item_help, '\n'])

    def start_section(self, heading):
        self._indent()
        section = self._Section(self, self._current_section, heading)
        self._add_item(section.format_help, [])
        self._current_section = section

    def _get_help_string(self, action):
        help_string = ''
        if '%(default)' not in action.help:
            if action.default is not argparse.SUPPRESS:
                from jina.helper import colored

                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if isinstance(action, argparse._StoreTrueAction):

                    help_string = colored(
                        'default: %s'
                        % (
                            'enabled'
                            if action.default
                            else f'disabled, use "{action.option_strings[0]}" to enable it'
                        ),
                        attrs=['dark'],
                    )
                elif action.choices:
                    choices_str = f'{{{", ".join([str(c) for c in action.choices])}}}'
                    help_string = colored(
                        'choose from: ' + choices_str + '; default: %(default)s',
                        attrs=['dark'],
                    )
                elif action.option_strings or action.nargs in defaulting_nargs:
                    help_string = colored(
                        'type: %(type)s; default: %(default)s', attrs=['dark']
                    )
        return f'''
        
        {help_string}
        
        {action.help}
        
        '''

    def _join_parts(self, part_strings):
        return '\n' + ''.join(
            [part for part in part_strings if part and part is not argparse.SUPPRESS]
        )

    def _get_default_metavar_for_optional(self, action):
        return ''

    def _expand_help(self, action):
        params = dict(vars(action), prog=self._prog)
        for name in list(params):
            if params[name] is argparse.SUPPRESS:
                del params[name]
        for name in list(params):
            if hasattr(params[name], '__name__'):
                params[name] = params[name].__name__
        return self._get_help_string(action) % params

    def _metavar_formatter(self, action, default_metavar):
        if action.metavar is not None:
            result = action.metavar
        elif action.choices is not None:

            if len(action.choices) > 4:
                choice_strs = ', '.join([str(c) for c in action.choices][:4])
                result = f'{{{choice_strs} ... {len(action.choices) - 4} more choices}}'
            else:
                choice_strs = ', '.join([str(c) for c in action.choices])
                result = f'{{{choice_strs}}}'
        else:
            result = default_metavar

        def formatter(tuple_size):
            if isinstance(result, tuple):
                return result
            else:
                return (result,) * tuple_size

        return formatter

    def _split_lines(self, text, width):
        return self._para_reformat(text, width)

    def _fill_text(self, text, width, indent):
        lines = self._para_reformat(text, width)
        return '\n'.join(lines)

    def _indents(self, line) -> Tuple[int, int]:
        """Return line indent level and "sub_indent" for bullet list text.

        :param line: the line to check
        :return: indentation of line and indentation of sub-items
        """
        import re

        indent = len(re.match(r'( *)', line).group(1))
        list_match = re.match(r'( *)(([*\-+>]+|\w+\)|\w+\.) +)', line)
        if list_match:
            sub_indent = indent + len(list_match.group(2))
        else:
            sub_indent = indent

        return indent, sub_indent

    def _split_paragraphs(self, text):
        """Split text into paragraphs of like-indented lines.

        :param text: the text input
        :return: list of paragraphs
        """

        import re
        import textwrap

        text = textwrap.dedent(text).strip()
        text = re.sub('\n\n[\n]+', '\n\n', text)

        last_sub_indent = None
        paragraphs = list()
        for line in text.splitlines():
            (indent, sub_indent) = self._indents(line)
            is_text = len(line.strip()) > 0

            if is_text and indent == sub_indent == last_sub_indent:
                paragraphs[-1] += ' ' + line
            else:
                paragraphs.append(line)

            if is_text:
                last_sub_indent = sub_indent
            else:
                last_sub_indent = None

        return paragraphs

    def _para_reformat(self, text, width):
        """Format text, by paragraph.

        :param text: the text to format
        :param width: the width to apply
        :return: the new text
        """

        import textwrap

        lines = list()
        for paragraph in self._split_paragraphs(text):
            (indent, sub_indent) = self._indents(paragraph)

            paragraph = self._whitespace_matcher.sub(' ', paragraph).strip()
            new_lines = textwrap.wrap(
                text=paragraph,
                width=width,
                initial_indent=' ' * indent,
                subsequent_indent=' ' * sub_indent,
            )

            # Blank lines get eaten by textwrap, put it back
            lines.extend(new_lines or [''])

        return lines


def _get_gateway_class(protocol, works_as_load_balancer = False):
    from jina.serve.runtimes.gateway.grpc import GRPCGateway
    from jina.serve.runtimes.gateway.http import HTTPGateway
    from jina.serve.runtimes.gateway.websocket import WebSocketGateway

    gateway_dict = {
        ProtocolType.GRPC: GRPCGateway,
        ProtocolType.WEBSOCKET: WebSocketGateway,
        ProtocolType.HTTP: HTTPGateway,
    }
    if protocol == ProtocolType.HTTP and works_as_load_balancer:
        from jina.serve.runtimes.gateway.load_balancer import LoadBalancerGateway
        return LoadBalancerGateway
    else:
        return gateway_dict[protocol]


def _set_gateway_uses(args: 'argparse.Namespace', gateway_load_balancer: bool = False):
    if not args.uses:
        if len(args.protocol) == 1 and len(args.port) == 1:
            args.uses = _get_gateway_class(args.protocol[0]).__name__
        elif len(args.protocol) > len(args.port):
            if len(args.port) == 1:
                from jina.helper import random_port
                args.port = []
                for _ in range(len(args.protocol)):
                    args.port.append(random_port())
            else:
                raise ValueError(
                    'You need to specify as much protocols as ports if you want to use a jina built-in gateway'
                )
        if len(args.protocol) > 1:
            from jina.serve.runtimes.gateway.composite import CompositeGateway
            args.uses = CompositeGateway.__name__
        elif gateway_load_balancer:
            from jina.serve.runtimes.gateway.load_balancer import LoadBalancerGateway
            args.uses = LoadBalancerGateway.__name__


def _update_gateway_args(args: 'argparse.Namespace', **kwargs):
    _set_gateway_uses(args, **kwargs)


class CastToIntAction(argparse.Action):
    """argparse action to cast a list of values to int"""

    def __call__(self, parser, args, values, option_string=None):
        """
        call the CastToIntAction


        .. # noqa: DAR401
        :param parser: the parser
        :param args: args to initialize the values
        :param values: the values to add to the parser
        :param option_string: inherited, not used
        """
        d = []
        for value in values:
            value = value.split(',')
            d.extend([_port_to_int(port) for port in value])
        setattr(args, self.dest, d)


class CastPeerPorts(argparse.Action):
    """argparse action to cast potential inputs to `peer-ports` argument"""

    def __call__(self, parser, args, values, option_string=None):
        """
        call the CastPeerPorts


        .. # noqa: DAR401
        :param parser: the parser
        :param args: args to initialize the values
        :param values: the values to add to the parser
        :param option_string: inherited, not used
        """
        import json
        d = {0: []}
        for value in values:
            if isinstance(value, str):
                value = json.loads(value)
            if isinstance(value, dict):
                for k, vlist in value.items():
                    d[k] = []
                    for v in vlist:
                        d[k].append(_port_to_int(v))
            elif isinstance(value, int):
                d[0].append(value)
            else:
                d[0] = [_port_to_int(port) for port in value]
        setattr(args, self.dest, d)


def _port_to_int(port):
    try:
        return int(port)
    except ValueError:
        default_logger.warning(
            f'port {port} is not an integer and cannot be cast to one'
        )
        return port


class CastHostAction(argparse.Action):
    """argparse action to cast a list of values to int"""

    def __call__(self, parser, args, values, option_string=None):
        """
        call the CastHostAction


        .. # noqa: DAR401
        :param parser: the parser
        :param args: args to initialize the values
        :param values: the values to add to the parser
        :param option_string: inherited, not used
        """
        d = []
        for value in values:
            d.extend(value.split(','))
        setattr(args, self.dest, d)


_chf = _ColoredHelpFormatter



from google.protobuf import __version__ as __pb__version__

if __pb__version__.startswith('4'):
    from jina.serve.consensus.add_voter.pb.add_voter_pb2_grpc import *
else:
    from jina.serve.consensus.add_voter.pb2.add_voter_pb2_grpc import *



import argparse
import multiprocessing
import os
import copy
from typing import Dict, Optional, Type, Union, TYPE_CHECKING

from jina.logging.logger import JinaLogger
from jina.serve.runtimes.asyncio import AsyncNewLoopRuntime
from jina.serve.helper import _get_workspace_from_name_and_shards
from jina.constants import RAFT_TO_EXECUTOR_PORT

if TYPE_CHECKING:
    import threading


def run_raft(
        args: 'argparse.Namespace',
        is_ready: Union['multiprocessing.Event', 'threading.Event'],
):
    """Method to run the RAFT

    This method is the target for the Pod's `thread` or `process`

    :param args: namespace args from the Pod
    :param is_ready: concurrency event to communicate Executor runtime is ready to receive messages
    """

    import jraft

    def pascal_case_dict(d):
        new_d = {}
        for key, value in d.items():
            new_key = key
            if '_' in key:
                new_key = ''.join(word.capitalize() for word in key.split('_'))
            new_d[new_key] = value
        return new_d

    raft_id = str(args.replica_id)
    shard_id = args.shard_id if args.shards > 1 else -1

    raft_dir = _get_workspace_from_name_and_shards(
        workspace=args.workspace, name='raft', shard_id=shard_id
    )

    port = args.port[0] if isinstance(args.port, list) else args.port
    address = f'{args.host}:{port}'
    executor_target = f'{args.host}:{port + RAFT_TO_EXECUTOR_PORT}'

    # if the Executor was already persisted, retrieve its port and host configuration
    logger = JinaLogger(context=f'RAFT-{args.name}', **vars(args))
    persisted_address = jraft.get_configuration(raft_id, raft_dir)
    if persisted_address:
        logger.debug(f'Configuration found on the node: Address {persisted_address}')
        address = persisted_address
        executor_host, port = persisted_address.split(':')
        executor_target = f'{executor_host}:{int(port) + 1}'

    raft_configuration = pascal_case_dict(args.raft_configuration or {})
    log_level = raft_configuration.get('LogLevel', os.getenv('JINA_LOG_LEVEL', 'INFO'))
    raft_configuration['LogLevel'] = log_level
    is_ready.wait()
    logger.debug(f'Will run the RAFT node with RAFT configuration {raft_configuration}')
    jraft.run(
        address,
        raft_id,
        raft_dir,
        args.name,
        executor_target,
        **raft_configuration,
    )


def run(
        args: 'argparse.Namespace',
        name: str,
        runtime_cls: Type[AsyncNewLoopRuntime],
        envs: Dict[str, str],
        is_started: Union['multiprocessing.Event', 'threading.Event'],
        is_shutdown: Union['multiprocessing.Event', 'threading.Event'],
        is_ready: Union['multiprocessing.Event', 'threading.Event'],
        is_signal_handlers_installed: Union['multiprocessing.Event', 'threading.Event'],
        jaml_classes: Optional[Dict] = None,
):
    """Method representing the :class:`BaseRuntime` activity.

    This method is the target for the Pod's `thread` or `process`

    .. note::
        :meth:`run` is running in subprocess/thread, the exception can not be propagated to the main process.
        Hence, please do not raise any exception here.

    .. note::
        Please note that env variables are process-specific. Subprocess inherits envs from
        the main process. But Subprocess's envs do NOT affect the main process. It does NOT
        mess up user local system envs.

    .. warning::
        If you are using ``thread`` as backend, envs setting will likely be overidden by others

    .. note::
        `jaml_classes` contains all the :class:`JAMLCompatible` classes registered in the main process.
        When using `spawn` as the multiprocessing start method, passing this argument to `run` method re-imports
        & re-registers all `JAMLCompatible` classes.

    :param args: namespace args from the Pod
    :param name: name of the Pod to have proper logging
    :param runtime_cls: the runtime class to instantiate
    :param envs: a dictionary of environment variables to be set in the new Process
    :param is_started: concurrency event to communicate runtime is properly started. Used for better logging
    :param is_shutdown: concurrency event to communicate runtime is terminated
    :param is_ready: concurrency event to communicate runtime is ready to receive messages
    :param is_signal_handlers_installed: concurrency event to communicate runtime is ready to get SIGTERM from orchestration
    :param jaml_classes: all the `JAMLCompatible` classes imported in main process
    """
    req_handler_cls = None
    if runtime_cls == 'GatewayRuntime':
        from jina.serve.runtimes.gateway.request_handling import GatewayRequestHandler
        req_handler_cls = GatewayRequestHandler
    elif runtime_cls == 'WorkerRuntime':
        from jina.serve.runtimes.worker.request_handling import WorkerRequestHandler
        req_handler_cls = WorkerRequestHandler
    elif runtime_cls == 'HeadRuntime':
        from jina.serve.runtimes.head.request_handling import HeaderRequestHandler
        req_handler_cls = HeaderRequestHandler

    logger = JinaLogger(name, **vars(args))

    def _unset_envs():
        if envs:
            for k in envs.keys():
                os.environ.pop(k, None)

    def _set_envs():
        if args.env:
            os.environ.update({k: str(v) for k, v in envs.items()})

    try:
        _set_envs()

        runtime = AsyncNewLoopRuntime(
            args=args,
            req_handler_cls=req_handler_cls,
            gateway_load_balancer=getattr(args, 'gateway_load_balancer', False),
            signal_handlers_installed_event=is_signal_handlers_installed
        )
    except Exception as ex:
        logger.error(
            f'{ex!r} during {runtime_cls!r} initialization'
            + f'\n add "--quiet-error" to suppress the exception details'
            if not args.quiet_error
            else '',
            exc_info=not args.quiet_error,
        )
    else:
        if not is_shutdown.is_set():
            is_started.set()
            with runtime:
                # here the ready event is being set
                is_ready.set()
                runtime.run_forever()
    finally:
        _unset_envs()
        is_shutdown.set()
        logger.debug('process terminated')


def run_stateful(args: 'argparse.Namespace',
                 name: str,
                 runtime_cls: Type[AsyncNewLoopRuntime],
                 envs: Dict[str, str]):
    """
    Method to be called in Docker containers when Stateful Executor is required. This will start
    2 processes in the Docker container.
    :param args: namespace args from the Pod
    :param name: name of the Pod to have proper logging
    :param runtime_cls: the runtime class to instantiate
    :param envs: a dictionary of environment variables to be set in the new Process
    """
    import signal
    from jina.jaml import JAML
    is_ready = multiprocessing.Event()
    is_shutdown = multiprocessing.Event()
    is_started = multiprocessing.Event()
    is_signal_handlers_installed = multiprocessing.Event()
    raft_worker = multiprocessing.Process(
        target=run_raft,
        kwargs={
            'args': args,
            'is_ready': is_ready,
        },
        name=name,
        daemon=True,
    )
    cargs = copy.deepcopy(args)

    from jina.constants import RAFT_TO_EXECUTOR_PORT

    if isinstance(cargs.port, int):
        cargs.port += RAFT_TO_EXECUTOR_PORT
    elif isinstance(cargs.port, list):
        cargs.port = [port + RAFT_TO_EXECUTOR_PORT for port in cargs.port]
    worker = multiprocessing.Process(
        target=run,
        kwargs={
            'args': cargs,
            'name': name,
            'envs': envs,
            'is_started': is_started,
            'is_shutdown': is_shutdown,
            'is_ready': is_ready,
            'is_signal_handlers_installed': is_signal_handlers_installed,
            'runtime_cls': runtime_cls,
            'jaml_classes': JAML.registered_classes(),
        },
        name=name,
        daemon=False,
    )

    try:
        HANDLED_SIGNALS = (
            signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
            signal.SIGTERM,  # Unix signal 15. Sent by `kill <pid>`.
            signal.SIGSEGV,
        )

        def signal_handler(*args, **kwargs):
            worker.terminate()
            raft_worker.terminate()

        for sig in HANDLED_SIGNALS:
            signal.signal(sig, signal_handler)

        worker.start()
        raft_worker.start()

        worker.join()
        raft_worker.join()
    except:
        worker.terminate()
        raft_worker.terminate()



import asyncio
import json
import os
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from jina._docarray import Document, DocumentArray, docarray_v2
from jina.excepts import ExecutorError
from jina.logging.logger import JinaLogger
from jina.proto import jina_pb2
from jina.serve.networking import GrpcConnectionPool
from jina.serve.runtimes.gateway.async_request_response_handling import (
    AsyncRequestResponseHandler,
)
from jina.serve.runtimes.gateway.graph.topology_graph import TopologyGraph
from jina.serve.stream import RequestStreamer
from jina.types.request import Request
from jina.types.request.data import DataRequest, SingleDocumentRequest

if docarray_v2:
    from docarray import DocList

__all__ = ['GatewayStreamer']

if TYPE_CHECKING:  # pragma: no cover
    from grpc.aio._interceptor import ClientInterceptor
    from opentelemetry.instrumentation.grpc._client import (
        OpenTelemetryClientInterceptor,
    )
    from opentelemetry.metrics import Meter
    from prometheus_client import CollectorRegistry


class GatewayStreamer:
    """
    Wrapper object to be used in a Custom Gateway. Naming to be defined
    """

    def __init__(
        self,
        graph_representation: Dict,
        executor_addresses: Dict[str, Union[str, List[str]]],
        graph_conditions: Dict = {},
        deployments_metadata: Dict[str, Dict[str, str]] = {},
        deployments_no_reduce: List[str] = [],
        timeout_send: Optional[float] = None,
        retries: int = 0,
        compression: Optional[str] = None,
        runtime_name: str = 'custom gateway',
        prefetch: int = 0,
        logger: Optional['JinaLogger'] = None,
        metrics_registry: Optional['CollectorRegistry'] = None,
        meter: Optional['Meter'] = None,
        aio_tracing_client_interceptors: Optional[Sequence['ClientInterceptor']] = None,
        tracing_client_interceptor: Optional['OpenTelemetryClientInterceptor'] = None,
        grpc_channel_options: Optional[list] = None,
    ):
        """
        :param graph_representation: A dictionary describing the topology of the Deployments. 2 special nodes are expected, the name `start-gateway` and `end-gateway` to
            determine the nodes that receive the very first request and the ones whose response needs to be sent back to the client. All the nodes with no outgoing nodes
            will be considered to be floating, and they will be "flagged" so that the user can ignore their tasks and not await them.
        :param executor_addresses: dictionary JSON with the input addresses of each Deployment. Each Executor can have one single address or a list of addrresses for each Executor
        :param graph_conditions: Dictionary stating which filtering conditions each Executor in the graph requires to receive Documents.
        :param deployments_metadata: Dictionary with the metadata of each Deployment. Each executor deployment can have a list of key-value pairs to
            provide information associated with the request to the deployment.
        :param deployments_no_reduce: list of Executor disabling the built-in merging mechanism.
        :param timeout_send: Timeout to be considered when sending requests to Executors
        :param retries: Number of retries to try to make successfull sendings to Executors
        :param compression: The compression mechanism used when sending requests from the Head to the WorkerRuntimes. For more details, check https://grpc.github.io/grpc/python/grpc.html#compression.
        :param runtime_name: Name to be used for monitoring.
        :param prefetch: How many Requests are processed from the Client at the same time.
        :param logger: Optional logger that can be used for logging
        :param metrics_registry: optional metrics registry for prometheus used if we need to expose metrics
        :param meter: optional OpenTelemetry meter that can provide instruments for collecting metrics
        :param aio_tracing_client_interceptors: Optional list of aio grpc tracing server interceptors.
        :param tracing_client_interceptor: Optional gprc tracing server interceptor.
        :param grpc_channel_options: Optional gprc channel options.
        """
        self.logger = logger or JinaLogger(self.__class__.__name__)
        self.topology_graph = TopologyGraph(
            graph_representation=graph_representation,
            graph_conditions=graph_conditions,
            deployments_metadata=deployments_metadata,
            deployments_no_reduce=deployments_no_reduce,
            timeout_send=timeout_send,
            retries=retries,
            logger=logger,
        )

        self.runtime_name = runtime_name
        self.aio_tracing_client_interceptors = aio_tracing_client_interceptors
        self.tracing_client_interceptor = tracing_client_interceptor
        self._executor_addresses = executor_addresses

        self._connection_pool = self._create_connection_pool(
            executor_addresses,
            compression,
            metrics_registry,
            meter,
            logger,
            aio_tracing_client_interceptors,
            tracing_client_interceptor,
            grpc_channel_options,
        )
        request_handler = AsyncRequestResponseHandler(
            metrics_registry, meter, runtime_name, logger
        )
        self._single_doc_request_handler = (
            request_handler.handle_single_document_request(
                graph=self.topology_graph, connection_pool=self._connection_pool
            )
        )
        self._streamer = RequestStreamer(
            request_handler=request_handler.handle_request(
                graph=self.topology_graph, connection_pool=self._connection_pool
            ),
            result_handler=request_handler.handle_result(),
            prefetch=prefetch,
            logger=logger,
        )
        self._endpoints_models_map = None
        self._streamer.Call = self._streamer.stream

    def _create_connection_pool(
        self,
        deployments_addresses,
        compression,
        metrics_registry,
        meter,
        logger,
        aio_tracing_client_interceptors,
        tracing_client_interceptor,
        grpc_channel_options=None,
    ):
        # add the connections needed
        connection_pool = GrpcConnectionPool(
            runtime_name=self.runtime_name,
            logger=logger,
            compression=compression,
            metrics_registry=metrics_registry,
            meter=meter,
            aio_tracing_client_interceptors=aio_tracing_client_interceptors,
            tracing_client_interceptor=tracing_client_interceptor,
            channel_options=grpc_channel_options,
        )
        for deployment_name, addresses in deployments_addresses.items():
            for address in addresses:
                connection_pool.add_connection(
                    deployment=deployment_name, address=address, head=True
                )

        return connection_pool

    def rpc_stream(self, *args, **kwargs):
        """
        stream requests from client iterator and stream responses back.

        :param args: positional arguments to be passed to inner RequestStreamer
        :param kwargs: keyword arguments to be passed to inner RequestStreamer
        :return: An iterator over the responses from the Executors
        """
        return self._streamer.stream(*args, **kwargs)

    def rpc_stream_doc(self, *args, **kwargs):
        """
        stream requests from client iterator and stream responses back.

        :param args: positional arguments to be passed to inner RequestStreamer
        :param kwargs: keyword arguments to be passed to inner RequestStreamer
        :return: An iterator over the responses from the Executors
        """
        return self._single_doc_request_handler(*args, **kwargs)

    async def _get_endpoints_input_output_models(self, is_cancel):
        """
        Return a Dictionary with endpoints as keys and values as a dictionary of input and output schemas and names
        taken from the endpoints proto endpoint of Executors.
        :param is_cancel: event signal to show that you should stop trying
        """
        # The logic should be to get the response of all the endpoints protos schemas from all the nodes. Then do a
        # logic that for every endpoint fom every Executor computes what is the input and output schema seen by the
        # Flow.
        self._endpoints_models_map = (
            await self._streamer._get_endpoints_input_output_models(
                self.topology_graph, self._connection_pool, is_cancel
            )
        )

    def _validate_flow_docarray_compatibility(self):
        """
        This method aims to validate that the input-output docarray models of Executors are good
        """
        self.topology_graph._validate_flow_docarray_compatibility()

    async def stream(
        self,
        docs: DocumentArray,
        request_size: int = 100,
        return_results: bool = False,
        exec_endpoint: Optional[str] = None,
        target_executor: Optional[str] = None,
        parameters: Optional[Dict] = None,
        results_in_order: bool = False,
        return_type: Type[DocumentArray] = DocumentArray,
    ) -> AsyncIterator[Tuple[Union[DocumentArray, 'Request'], 'ExecutorError']]:
        """
        stream Documents and yield Documents or Responses and unpacked Executor error if any.

        :param docs: The Documents to be sent to all the Executors
        :param request_size: The amount of Documents to be put inside a single request.
        :param return_results: If set to True, the generator will yield Responses and not `DocumentArrays`
        :param exec_endpoint: The Executor endpoint to which to send the Documents
        :param target_executor: A regex expression indicating the Executors that should receive the Request
        :param parameters: Parameters to be attached to the Requests
        :param results_in_order: return the results in the same order as the request_iterator
        :param return_type: the DocumentArray type to be returned. By default, it is `DocumentArray`.
        :yield: tuple of Documents or Responses and unpacked error from Executors if any
        """
        async for result in self.stream_docs(
            docs=docs,
            request_size=request_size,
            return_results=True,  # force return Responses
            exec_endpoint=exec_endpoint,
            target_executor=target_executor,
            parameters=parameters,
            results_in_order=results_in_order,
            return_type=return_type,
        ):
            error = None
            if jina_pb2.StatusProto.ERROR == result.status.code:
                exception = result.status.exception
                error = ExecutorError(
                    name=exception.name,
                    args=exception.args,
                    stacks=exception.stacks,
                    executor=exception.executor,
                )
            if return_results:
                yield result, error
            else:
                yield result.data.docs, error

    async def stream_doc(
        self,
        doc: 'Document',
        return_results: bool = False,
        exec_endpoint: Optional[str] = None,
        target_executor: Optional[str] = None,
        parameters: Optional[Dict] = None,
        request_id: Optional[str] = None,
        return_type: Type[DocumentArray] = DocumentArray,
    ) -> AsyncIterator[Tuple[Union[DocumentArray, 'Request'], 'ExecutorError']]:
        """
        stream Documents and yield Documents or Responses and unpacked Executor error if any.

        :param doc: The Documents to be sent to all the Executors
        :param return_results: If set to True, the generator will yield Responses and not `DocumentArrays`
        :param exec_endpoint: The Executor endpoint to which to send the Documents
        :param target_executor: A regex expression indicating the Executors that should receive the Request
        :param parameters: Parameters to be attached to the Requests
        :param request_id: Request ID to add to the request streamed to Executor. Only applicable if request_size is equal or less to the length of the docs
        :param return_type: the DocumentArray type to be returned. By default, it is `DocumentArray`.
        :yield: tuple of Documents or Responses and unpacked error from Executors if any
        """
        req = SingleDocumentRequest()
        req.document_cls = doc.__class__
        req.data.doc = doc
        if request_id:
            req.header.request_id = request_id
        if exec_endpoint:
            req.header.exec_endpoint = exec_endpoint
        if target_executor:
            req.header.target_executor = target_executor
        if parameters:
            req.parameters = parameters

        async for result in self.rpc_stream_doc(request=req, return_type=return_type):
            error = None
            if jina_pb2.StatusProto.ERROR == result.status.code:
                exception = result.status.exception
                error = ExecutorError(
                    name=exception.name,
                    args=exception.args,
                    stacks=exception.stacks,
                    executor=exception.executor,
                )
            if return_results:
                yield result, error
            else:
                yield result.data.doc, error

    async def stream_docs(
        self,
        docs: DocumentArray,
        request_size: int = 100,
        return_results: bool = False,
        exec_endpoint: Optional[str] = None,
        target_executor: Optional[str] = None,
        parameters: Optional[Dict] = None,
        results_in_order: bool = False,
        request_id: Optional[str] = None,
        return_type: Type[DocumentArray] = DocumentArray,
    ):
        """
        stream documents and stream responses back.

        :param docs: The Documents to be sent to all the Executors
        :param request_size: The amount of Documents to be put inside a single request.
        :param return_results: If set to True, the generator will yield Responses and not `DocumentArrays`
        :param exec_endpoint: The Executor endpoint to which to send the Documents
        :param target_executor: A regex expression indicating the Executors that should receive the Request
        :param parameters: Parameters to be attached to the Requests
        :param results_in_order: return the results in the same order as the request_iterator
        :param request_id: Request ID to add to the request streamed to Executor. Only applicable if request_size is equal or less to the length of the docs
        :param return_type: the DocumentArray type to be returned. By default, it is `DocumentArray`.
        :yield: Yields DocumentArrays or Responses from the Executors
        """
        request_id = request_id if len(docs) <= request_size else None

        def _req_generator():
            if not docarray_v2:
                for docs_batch in docs.batch(batch_size=request_size, shuffle=False):
                    req = DataRequest()
                    req.data.docs = docs_batch
                    if request_id:
                        req.header.request_id = request_id
                    if exec_endpoint:
                        req.header.exec_endpoint = exec_endpoint
                    if target_executor:
                        req.header.target_executor = target_executor
                    if parameters:
                        req.parameters = parameters
                    yield req
            else:
                from docarray import BaseDoc

                def batch(iterable, n=1):
                    l = len(iterable)
                    for ndx in range(0, l, n):
                        yield iterable[ndx : min(ndx + n, l)]

                if len(docs) > 0:
                    for docs_batch in batch(docs, n=request_size):
                        req = DataRequest()
                        req.document_array_cls = DocList[docs_batch.doc_type]
                        req.data.docs = docs_batch
                        if request_id:
                            req.header.request_id = request_id
                        if exec_endpoint:
                            req.header.exec_endpoint = exec_endpoint
                        if target_executor:
                            req.header.target_executor = target_executor
                        if parameters:
                            req.parameters = parameters
                        yield req
                else:
                    req = DataRequest()
                    req.document_array_cls = DocList[BaseDoc]
                    req.data.docs = DocList[BaseDoc]()
                    if request_id:
                        req.header.request_id = request_id
                    if exec_endpoint:
                        req.header.exec_endpoint = exec_endpoint
                    if target_executor:
                        req.header.target_executor = target_executor
                    if parameters:
                        req.parameters = parameters
                    yield req

        async for resp in self.rpc_stream(
            request_iterator=_req_generator(),
            results_in_order=results_in_order,
            return_type=return_type,
        ):
            if return_results:
                yield resp
            else:
                yield resp.docs

    async def close(self):
        """
        Gratefully closes the object making sure all the floating requests are taken care and the connections are closed gracefully
        """
        await self._streamer.wait_floating_requests_end()
        await self._connection_pool.close()

    Call = rpc_stream

    async def process_single_data(
        self, request: DataRequest, context=None
    ) -> DataRequest:
        """Implements request and response handling of a single DataRequest
        :param request: DataRequest from Client
        :param context: grpc context
        :return: response DataRequest
        """
        return await self._streamer.process_single_data(request, context)

    @staticmethod
    def get_streamer():
        """
        Return a streamer object based on the current environment context.
        The streamer object is contructed using runtime arguments stored in the `JINA_STREAMER_ARGS` environment variable.
        If this method is used outside a Jina context (process not controlled/orchestrated by jina), this method will
        raise an error.
        The streamer object does not have tracing/instrumentation capabilities.

        :return: Returns an instance of `GatewayStreamer`
        """
        if 'JINA_STREAMER_ARGS' in os.environ:
            args_dict = json.loads(os.environ['JINA_STREAMER_ARGS'])
            return GatewayStreamer(**args_dict)
        else:
            raise OSError('JINA_STREAMER_ARGS environment variable is not set')

    @staticmethod
    def _set_env_streamer_args(**kwargs):
        os.environ['JINA_STREAMER_ARGS'] = json.dumps(kwargs)


class _ExecutorStreamer:
    def __init__(self, connection_pool: GrpcConnectionPool, executor_name: str) -> None:
        self._connection_pool: GrpcConnectionPool = connection_pool
        self.executor_name = executor_name

    async def post(
        self,
        inputs: DocumentArray,
        request_size: int = 100,
        on: Optional[str] = None,
        parameters: Optional[Dict] = None,
        return_type: Type[DocumentArray] = DocumentArray,
        **kwargs,
    ):
        if not parameters:
            parameters = {}
        if not docarray_v2:
            reqs = []
            for docs_batch in inputs.batch(batch_size=request_size, shuffle=False):
                req = DataRequest()
                req.header.exec_endpoint = on
                req.header.target_executor = self.executor_name
                req.parameters = parameters
                req.data.docs = docs_batch
                reqs.append(req)
        else:
            from docarray import BaseDoc

            def batch(iterable, n=1):
                l = len(iterable)
                for ndx in range(0, l, n):
                    yield iterable[ndx : min(ndx + n, l)]

            reqs = []

            if len(inputs) > 0:
                for docs_batch in batch(inputs, n=request_size):
                    req = DataRequest()
                    req.document_array_cls = DocList[docs_batch.doc_type]
                    req.data.docs = docs_batch
                    req.header.exec_endpoint = on
                    req.header.target_executor = self.executor_name
                    req.parameters = parameters
                    reqs.append(req)
            else:
                req = DataRequest()
                req.document_array_cls = DocList[BaseDoc]
                req.data.docs = DocList[BaseDoc]()
                req.header.exec_endpoint = on
                req.header.target_executor = self.executor_name
                req.parameters = parameters
                reqs.append(req)

        tasks = [
            self._connection_pool.send_requests_once(
                requests=[req], deployment=self.executor_name, head=True, endpoint=on
            )
            for req in reqs
        ]

        results = await asyncio.gather(*tasks)

        if not docarray_v2:
            docs = DocumentArray.empty()
            for resp, _ in results:
                docs.extend(resp.docs)
        else:
            docs = DocList[return_type.doc_type]()
            for resp, _ in results:
                resp.document_array_cls = return_type
                docs.extend(resp.docs)
        return docs

    async def stream_doc(
        self,
        inputs: 'Document',
        on: Optional[str] = None,
        parameters: Optional[Dict] = None,
        **kwargs,
    ):
        req: SingleDocumentRequest = SingleDocumentRequest(inputs.to_protobuf())
        req.header.exec_endpoint = on
        req.header.target_executor = self.executor_name
        req.parameters = parameters
        async_generator = self._connection_pool.send_single_document_request(
            request=req, deployment=self.executor_name, head=True, endpoint=on
        )

        async for resp, _ in async_generator:
            yield resp
